# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from contextlib import contextmanager
import os.path

import eventlet
from neutron_lib import context as ctx
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall

from networking_l2gw.services.l2gateway.agent import base_agent_manager
from networking_l2gw.services.l2gateway.agent import l2gateway_config
from networking_l2gw.services.l2gateway.agent.ovsdb import ovsdb_common_class
from networking_l2gw.services.l2gateway.agent.ovsdb import ovsdb_monitor
from networking_l2gw.services.l2gateway.agent.ovsdb import ovsdb_writer
from networking_l2gw.services.l2gateway.common import constants as n_const

LOG = logging.getLogger(__name__)


class OVSDBManager(base_agent_manager.BaseAgentManager):
    """OVSDB variant of agent manager.

       Listens to state change notifications from OVSDB servers and
       handles transactions (RPCs) destined to OVSDB servers.
    """
    def __init__(self, conf=None):
        super(OVSDBManager, self).__init__(conf)
        self._extract_ovsdb_config(conf)
        self.enable_manager = cfg.CONF.ovsdb.enable_manager
        periodic_interval = self.conf.ovsdb.periodic_interval
        if self.enable_manager:
            self.ovsdb_fd = None
            self._sock_open_connection()
            self.looping_task_ovsdb_states = (
                loopingcall.FixedIntervalLoopingCall(self._send_ovsdb_states))
            self.looping_task_ovsdb_states.start(interval=periodic_interval)
        else:
            self.looping_task = loopingcall.FixedIntervalLoopingCall(
                self._connect_to_ovsdb_server)
            self.looping_task.start(interval=periodic_interval)

    def _extract_ovsdb_config(self, conf):
        self.conf = conf or cfg.CONF
        ovsdb_hosts = self.conf.ovsdb.ovsdb_hosts
        if ovsdb_hosts != '':
            ovsdb_hosts = ovsdb_hosts.split(',')
            for host in ovsdb_hosts:
                self._process_ovsdb_host(host)
            # Ensure that max_connection_retries is less than
            # the periodic interval.
            if (self.conf.ovsdb.max_connection_retries >=
                    self.conf.ovsdb.periodic_interval):
                raise SystemExit("max_connection_retries should be "
                                 "less than periodic interval")

    def _process_ovsdb_host(self, host):
        try:
            host_splits = str(host).split(':')
            ovsdb_identifier = str(host_splits[0]).strip()
            ovsdb_conf = {n_const.OVSDB_IDENTIFIER: ovsdb_identifier,
                          'ovsdb_ip': str(host_splits[1]).strip(),
                          'ovsdb_port': str(host_splits[2]).strip()}
            priv_key_path = self.conf.ovsdb.l2_gw_agent_priv_key_base_path
            cert_path = self.conf.ovsdb.l2_gw_agent_cert_base_path
            ca_cert_path = self.conf.ovsdb.l2_gw_agent_ca_cert_base_path
            use_ssl = priv_key_path and cert_path and ca_cert_path
            if use_ssl:
                LOG.debug("ssl is enabled with priv_key_path %s, cert_path "
                          "%s, ca_cert_path %s", priv_key_path,
                          cert_path, ca_cert_path)
                priv_key_file = priv_key_path + "/" + ovsdb_identifier + ".key"
                cert_file = cert_path + "/" + ovsdb_identifier + ".cert"
                ca_cert_file = (ca_cert_path + "/" + ovsdb_identifier +
                                ".ca_cert")
                is_priv_key = os.path.isfile(priv_key_file)
                is_cert_file = os.path.isfile(cert_file)
                is_ca_cert_file = os.path.isfile(ca_cert_file)
                if not is_priv_key:
                    LOG.exception("Could not find private key in "
                                  "%(path)s dir, expecting in the "
                                  "file name %(file)s ",
                                  {'path': priv_key_path,
                                   'file': ovsdb_identifier + ".key"})
                if not is_cert_file:
                    LOG.exception("Could not find cert in %(path)s dir, "
                                  "expecting in the file name %(file)s",
                                  {'path': cert_path,
                                   'file': ovsdb_identifier + ".cert"})
                if not is_ca_cert_file:
                    LOG.exception("Could not find cacert in %(path)s "
                                  "dir, expecting in the file name "
                                  "%(file)s",
                                  {'path': ca_cert_path,
                                   'file': ovsdb_identifier + ".ca_cert"})
                ssl_ovsdb = {'use_ssl': True,
                             'private_key':
                                 "/".join([str(priv_key_path),
                                           '.'.join([str(host_splits[0]).
                                                     strip(),
                                                     'key'])]),
                             'certificate':
                                 "/".join([str(cert_path),
                                           '.'.join([str(host_splits[0]).
                                                     strip(), 'cert'])]),
                             'ca_cert':
                                 "/".join([str(ca_cert_path),
                                           '.'.join([str(host_splits[0]).
                                                     strip(), 'ca_cert'])])
                             }
                ovsdb_conf.update(ssl_ovsdb)
            LOG.debug("ovsdb_conf = %s", str(ovsdb_conf))
            gateway = l2gateway_config.L2GatewayConfig(ovsdb_conf)
            self.gateways[ovsdb_identifier] = gateway
        except Exception as ex:
            LOG.exception("Exception %(ex)s occurred while processing "
                          "host %(host)s", {'ex': ex, 'host': host})

    def _connect_to_ovsdb_server(self):
        """Initializes the connection to the OVSDB servers."""
        ovsdb_states = {}
        if self.gateways and self.l2gw_agent_type == n_const.MONITOR:
            for key in self.gateways.keys():
                gateway = self.gateways.get(key)
                ovsdb_fd = gateway.ovsdb_fd
                if not (ovsdb_fd and ovsdb_fd.connected):
                    LOG.debug("OVSDB server %s is disconnected",
                              str(gateway.ovsdb_ip))
                    try:
                        ovsdb_fd = ovsdb_monitor.OVSDBMonitor(
                            self.conf.ovsdb,
                            gateway,
                            self.agent_to_plugin_rpc)
                    except Exception:
                        ovsdb_states[key] = 'disconnected'
                        # Log a warning and continue so that it can be
                        # retried in the next iteration.
                        LOG.error("OVSDB server %s is not "
                                  "reachable", gateway.ovsdb_ip)
                        # Continue processing the next element in the list.
                        continue
                    gateway.ovsdb_fd = ovsdb_fd
                    try:
                        eventlet.greenthread.spawn_n(
                            ovsdb_fd.set_monitor_response_handler)
                    except Exception:
                        raise SystemExit(Exception.message)
                if ovsdb_fd and ovsdb_fd.connected:
                    ovsdb_states[key] = 'connected'
        LOG.debug("Calling notify_ovsdb_states")
        self.plugin_rpc.notify_ovsdb_states(ctx.get_admin_context(),
                                            ovsdb_states)

    def handle_report_state_failure(self):
        # Not able to deliver the heart beats to the Neutron server.
        # Let us change the mode to Transact so that when the
        # Neutron server is connected back, it will make an agent
        # Monitor agent and OVSDB data will be read entirely. This way,
        # the OVSDB data in Neutron database will be the latest and in
        # sync with that in the OVSDB server tables.
        if self.l2gw_agent_type == n_const.MONITOR:
            self.l2gw_agent_type = ''
            self.agent_state.get('configurations')[n_const.L2GW_AGENT_TYPE
                                                   ] = self.l2gw_agent_type
            if not self.enable_manager:
                self._stop_looping_task()
                self._disconnect_all_ovsdb_servers()

    def _send_ovsdb_states(self):
        self.plugin_rpc.notify_ovsdb_states(ctx.get_admin_context(),
                                            self.ovsdb_fd.ovsdb_fd_states)

    def _disconnect_all_ovsdb_servers(self):
        if self.gateways:
            for key, gateway in self.gateways.items():
                ovsdb_fd = gateway.ovsdb_fd
                if ovsdb_fd and ovsdb_fd.connected:
                    gateway.ovsdb_fd.disconnect()

    def set_monitor_agent(self, context, hostname):
        """Handle RPC call from plugin to update agent type.

        RPC call from the plugin to accept that I am a monitoring
        or a transact agent. This is a fanout cast message
        """
        super(OVSDBManager, self).set_monitor_agent(context, hostname)

        # If set to Monitor, then let us start monitoring the OVSDB
        # servers without any further delay.
        if ((self.l2gw_agent_type == n_const.MONITOR) and not (
                self.enable_manager)):
            self._start_looping_task()
        elif self.enable_manager and self.l2gw_agent_type == n_const.MONITOR:
            if self.ovsdb_fd is None:
                self._sock_open_connection()
            elif ((self.ovsdb_fd) and not (
                  self.ovsdb_fd.check_monitor_table_thread) and (
                  self.ovsdb_fd.check_sock_rcv)):
                for key in self.ovsdb_fd.ovsdb_dicts.keys():
                    if key in self.ovsdb_fd.ovsdb_conn_list:
                        eventlet.greenthread.spawn_n(
                            self.ovsdb_fd._spawn_monitor_table_thread, key)
                self._start_looping_task_ovsdb_states()
        elif ((self.enable_manager) and not (
              self.l2gw_agent_type == n_const.MONITOR)):
            self._stop_looping_task_ovsdb_states()
        elif (not (self.l2gw_agent_type == n_const.MONITOR) and not (
                self.enable_manager)):
            # Otherwise, stop monitoring the OVSDB servers
            # and close the open connections if any.
            self._stop_looping_task()
            self._disconnect_all_ovsdb_servers()

    def _stop_looping_task_ovsdb_states(self):
        if self.looping_task_ovsdb_states._running:
            self.looping_task_ovsdb_states.stop()

    def _start_looping_task_ovsdb_states(self):
        if not self.looping_task_ovsdb_states._running:
            self.looping_task_ovsdb_states.start(
                interval=self.conf.ovsdb.periodic_interval)

    def _stop_looping_task(self):
        if self.looping_task._running:
            self.looping_task.stop()

    def _start_looping_task(self):
        if not self.looping_task._running:
            self.looping_task.start(interval=self.conf.ovsdb.
                                    periodic_interval)

    def _sock_open_connection(self):
        gateway = ''
        if self.ovsdb_fd is None:
            self.ovsdb_fd = ovsdb_common_class.OVSDB_commom_class(
                self.conf.ovsdb,
                gateway,
                self.agent_to_plugin_rpc, self)

    @contextmanager
    def _open_connection(self, ovsdb_identifier):
        ovsdb_fd = None
        gateway = self.gateways.get(ovsdb_identifier)
        try:
            ovsdb_fd = ovsdb_writer.OVSDBWriter(self.conf.ovsdb,
                                                gateway)
            yield ovsdb_fd
        finally:
            if ovsdb_fd:
                ovsdb_fd.disconnect()

    def _is_valid_request(self, ovsdb_identifier):
        val_req = ovsdb_identifier and ovsdb_identifier in self.gateways.keys()
        if not val_req:
            LOG.warning(n_const.ERROR_DICT
                        [n_const.L2GW_INVALID_OVSDB_IDENTIFIER])
        return val_req

    def delete_network(self, context, ovsdb_identifier, logical_switch_uuid):
        """Handle RPC cast from plugin to delete a network."""
        if self.enable_manager and self.l2gw_agent_type == n_const.MONITOR:
            self.ovsdb_fd.delete_logical_switch(logical_switch_uuid,
                                                ovsdb_identifier, False)
        elif ((self.enable_manager) and (
                not self.l2gw_agent_type == n_const.MONITOR)):
            self._sock_open_connection()
            if ovsdb_identifier in self.ovsdb_fd.ovsdb_conn_list:
                self.ovsdb_fd.delete_logical_switch(logical_switch_uuid,
                                                    ovsdb_identifier,
                                                    False)
        elif not self.enable_manager:
            if self._is_valid_request(ovsdb_identifier):
                with self._open_connection(ovsdb_identifier) as ovsdb_fd:
                    ovsdb_fd.delete_logical_switch(logical_switch_uuid,
                                                   ovsdb_identifier)

    def add_vif_to_gateway(self, context, ovsdb_identifier,
                           logical_switch_dict, locator_dict,
                           mac_dict):
        """Handle RPC cast from plugin to insert neutron port MACs."""
        if self.enable_manager and self.l2gw_agent_type == n_const.MONITOR:
            self.ovsdb_fd.insert_ucast_macs_remote(logical_switch_dict,
                                                   locator_dict,
                                                   mac_dict, ovsdb_identifier,
                                                   False)
        elif ((self.enable_manager) and (
                not self.l2gw_agent_type == n_const.MONITOR)):
            self._sock_open_connection()
            if ovsdb_identifier in self.ovsdb_fd.ovsdb_conn_list:
                self.ovsdb_fd.insert_ucast_macs_remote(
                    logical_switch_dict,
                    locator_dict,
                    mac_dict, ovsdb_identifier, False)
        elif not self.enable_manager:
            if self._is_valid_request(ovsdb_identifier):
                with self._open_connection(ovsdb_identifier) as ovsdb_fd:
                    ovsdb_fd.insert_ucast_macs_remote(logical_switch_dict,
                                                      locator_dict,
                                                      mac_dict,
                                                      ovsdb_identifier)

    def delete_vif_from_gateway(self, context, ovsdb_identifier,
                                logical_switch_uuid, mac):
        """Handle RPC cast from plugin to delete neutron port MACs."""
        if self.enable_manager and self.l2gw_agent_type == n_const.MONITOR:
            self.ovsdb_fd.delete_ucast_macs_remote(logical_switch_uuid, mac,
                                                   ovsdb_identifier,
                                                   False)
        elif ((self.enable_manager) and (
                not self.l2gw_agent_type == n_const.MONITOR)):
            self._sock_open_connection()
            if ovsdb_identifier in self.ovsdb_fd.ovsdb_conn_list:
                self.ovsdb_fd.delete_ucast_macs_remote(
                    logical_switch_uuid,
                    mac, ovsdb_identifier, False)
        elif not self.enable_manager:
            if self._is_valid_request(ovsdb_identifier):
                with self._open_connection(ovsdb_identifier) as ovsdb_fd:
                    ovsdb_fd.delete_ucast_macs_remote(
                        logical_switch_uuid, mac, ovsdb_identifier)

    def update_vif_to_gateway(self, context, ovsdb_identifier,
                              locator_dict, mac_dict):
        """Handle RPC cast from plugin to update neutron port MACs.

        for VM migration.
        """
        if self.enable_manager and self.l2gw_agent_type == n_const.MONITOR:
            self.ovsdb_fd.update_ucast_macs_remote(locator_dict,
                                                   mac_dict, ovsdb_identifier,
                                                   False)
        elif ((self.enable_manager) and (
                not self.l2gw_agent_type == n_const.MONITOR)):
            self._sock_open_connection()
            if ovsdb_identifier in self.ovsdb_fd.ovsdb_conn_list:
                self.ovsdb_fd.update_ucast_macs_remote(locator_dict,
                                                       mac_dict,
                                                       ovsdb_identifier, False)
        elif not self.enable_manager:
            if self._is_valid_request(ovsdb_identifier):
                with self._open_connection(ovsdb_identifier) as ovsdb_fd:
                    ovsdb_fd.update_ucast_macs_remote(
                        locator_dict, mac_dict, ovsdb_identifier)

    def update_connection_to_gateway(self, context, ovsdb_identifier,
                                     logical_switch_dict, locator_dicts,
                                     mac_dicts, port_dicts, op_method):
        """Handle RPC cast from plugin.

        Handle RPC cast from plugin to connect/disconnect a network
        to/from an L2 gateway.
        """
        if self.enable_manager and self.l2gw_agent_type == n_const.MONITOR:
            self.ovsdb_fd.update_connection_to_gateway(logical_switch_dict,
                                                       locator_dicts,
                                                       mac_dicts,
                                                       port_dicts,
                                                       ovsdb_identifier,
                                                       op_method,
                                                       False)
        elif ((self.enable_manager) and (
                not self.l2gw_agent_type == n_const.MONITOR)):
            self._sock_open_connection()
            if ovsdb_identifier in self.ovsdb_fd.ovsdb_conn_list:
                self.ovsdb_fd.update_connection_to_gateway(
                    logical_switch_dict,
                    locator_dicts,
                    mac_dicts,
                    port_dicts, ovsdb_identifier, op_method, False)
        elif not self.enable_manager:
            if self._is_valid_request(ovsdb_identifier):
                with self._open_connection(ovsdb_identifier) as ovsdb_fd:
                    ovsdb_fd.update_connection_to_gateway(logical_switch_dict,
                                                          locator_dicts,
                                                          mac_dicts,
                                                          port_dicts,
                                                          ovsdb_identifier,
                                                          op_method)

    def agent_to_plugin_rpc(self, activity, ovsdb_data):
        self.plugin_rpc.update_ovsdb_changes(ctx.get_admin_context(),
                                             activity,
                                             ovsdb_data)



# Copyright 2015 OpenStack Foundation
# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from networking_l2gw.db.l2gateway import db_query
from networking_l2gw.db.l2gateway import l2gateway_models as models
from networking_l2gw.extensions import l2gateway
from networking_l2gw.extensions import l2gatewayconnection
from networking_l2gw.services.l2gateway.common import config
from networking_l2gw.services.l2gateway.common import constants
from networking_l2gw.services.l2gateway.common import l2gw_validators
from networking_l2gw.services.l2gateway import exceptions as l2gw_exc

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib.db import api as db_api
from neutron_lib.db import model_query
from neutron_lib import exceptions
from neutron_lib.plugins import directory
from oslo_log import log as logging
from oslo_utils import uuidutils
from sqlalchemy.orm import exc as sa_orm_exc

LOG = logging.getLogger(__name__)


class L2GatewayMixin(l2gateway.L2GatewayPluginBase,
                     db_query.L2GatewayCommonDbMixin,
                     l2gatewayconnection.L2GatewayConnectionPluginBase):
    """Class L2GatewayMixin for handling l2_gateway resource."""
    gateway_resource = constants.GATEWAY_RESOURCE_NAME
    connection_resource = constants.CONNECTION_RESOURCE_NAME
    config.register_l2gw_opts_helper()

    @db_api.retry_if_session_inactive()
    @db_api.CONTEXT_READER
    def _get_l2_gateway(self, context, gw_id):
        try:
            gw = model_query.get_by_id(context, models.L2Gateway, gw_id)
        except sa_orm_exc.NoResultFound:
            raise l2gw_exc.L2GatewayNotFound(gateway_id=gw_id)
        return gw

    @db_api.retry_if_session_inactive()
    @db_api.CONTEXT_READER
    def _get_l2_gateways(self, context):
        return model_query.get_collection(context, models.L2Gateway,
                                          dict_func=None)

    @db_api.retry_if_session_inactive()
    @db_api.CONTEXT_READER
    def _get_l2_gw_interfaces(self, context, id):
        return context.session.query(models.L2GatewayInterface).filter_by(
            device_id=id).all()

    @db_api.CONTEXT_READER
    def _is_vlan_configured_on_any_interface_for_l2gw(self,
                                                      context,
                                                      l2gw_id):
        devices_db = self._get_l2_gateway_devices(context, l2gw_id)
        for device_model in devices_db:
            interfaces_db = self._get_l2_gw_interfaces(context,
                                                       device_model.id)
            for int_model in interfaces_db:
                query = context.session.query(models.L2GatewayInterface)
                int_db = query.filter_by(id=int_model.id).first()
                seg_id = int_db[constants.SEG_ID]
                if seg_id > 0:
                    return True
        return False

    @db_api.retry_if_session_inactive()
    def _get_l2_gateway_devices(self, context, l2gw_id):
        return context.session.query(models.L2GatewayDevice).filter_by(
            l2_gateway_id=l2gw_id).all()

    def _get_l2gw_devices_by_name_andl2gwid(self, context, device_name,
                                            l2gw_id):
        return context.session.query(models.L2GatewayDevice).filter_by(
            device_name=device_name, l2_gateway_id=l2gw_id).all()

    def _get_l2_gateway_connection(self, context, cn_id):
        try:
            con = context.session.query(models.L2GatewayConnection).get(cn_id)
        except sa_orm_exc.NoResultFound:
            raise l2gw_exc.L2GatewayConnectionNotFound(id=cn_id)
        return con

    def _make_l2gw_connections_dict(self, gw_conn, fields=None):
        if gw_conn is None:
            raise l2gw_exc.L2GatewayConnectionNotFound(id="")
        segmentation_id = gw_conn['segmentation_id']
        if segmentation_id == 0:
            segmentation_id = ""
        res = {'id': gw_conn['id'],
               'network_id': gw_conn['network_id'],
               'l2_gateway_id': gw_conn['l2_gateway_id'],
               'tenant_id': gw_conn['tenant_id'],
               'segmentation_id': segmentation_id
               }
        return self._fields(res, fields)

    def _make_l2_gateway_dict(self, l2_gateway, fields=None):
        device_list = []
        for d in l2_gateway.devices:
            interface_list = []
            for interfaces_db in d.interfaces:
                seg_id = interfaces_db[constants.SEG_ID]
                if seg_id == 0:
                    seg_id = ""
                interface_list.append({'name':
                                       interfaces_db['interface_name'],
                                       constants.SEG_ID:
                                       seg_id})
            aligned_int__list = self._align_interfaces_list(interface_list)
            device_list.append({'device_name': d['device_name'],
                                'id': d['id'],
                                'interfaces': aligned_int__list})
        res = {'id': l2_gateway['id'],
               'name': l2_gateway['name'],
               'devices': device_list,
               'tenant_id': l2_gateway['tenant_id']}
        return self._fields(res, fields)

    def _set_mapping_info_defaults(self, mapping_info):
        if not mapping_info.get(constants.SEG_ID):
            mapping_info[constants.SEG_ID] = 0

    def _retrieve_gateway_connections(self, context, gateway_id,
                                      mapping_info={}, only_one=False):
        filters = {'l2_gateway_id': [gateway_id]}
        for k, v in mapping_info.items():
            if v and k != constants.SEG_ID:
                filters[k] = [v]
        query = self._get_collection_query(context,
                                           models.L2GatewayConnection,
                                           filters)
        return query.one() if only_one else query.all()

    def create_l2_gateway(self, context, l2_gateway):
        """Create a logical gateway."""
        self._admin_check(context, 'CREATE')
        gw = l2_gateway[self.gateway_resource]
        tenant_id = self._get_tenant_id_for_create(context, gw)
        devices = gw['devices']
        with db_api.CONTEXT_WRITER.using(context):
            gw_db = models.L2Gateway(
                id=gw.get('id', uuidutils.generate_uuid()),
                tenant_id=tenant_id,
                name=gw.get('name'))
            context.session.add(gw_db)
            l2gw_device_dict = {}
            for device in devices:
                l2gw_device_dict['l2_gateway_id'] = id
                device_name = device['device_name']
                l2gw_device_dict['device_name'] = device_name
                l2gw_device_dict['id'] = uuidutils.generate_uuid()
                uuid = self._generate_uuid()
                dev_db = models.L2GatewayDevice(id=uuid,
                                                l2_gateway_id=gw_db.id,
                                                device_name=device_name)
                context.session.add(dev_db)
                for interface_list in device['interfaces']:
                    int_name = interface_list.get('name')
                    if constants.SEG_ID in interface_list:
                        seg_id_list = interface_list.get(constants.SEG_ID)
                        for seg_ids in seg_id_list:
                            uuid = self._generate_uuid()
                            interface_db = self._get_int_model(uuid,
                                                               int_name,
                                                               dev_db.id,
                                                               seg_ids)
                            context.session.add(interface_db)
                    else:
                        uuid = self._generate_uuid()
                        interface_db = self._get_int_model(uuid,
                                                           int_name,
                                                           dev_db.id,
                                                           0)
                        context.session.add(interface_db)
                    context.session.query(models.L2GatewayDevice).all()
        return self._make_l2_gateway_dict(gw_db)

    @db_api.CONTEXT_WRITER
    def update_l2_gateway(self, context, id, l2_gateway):
        """Update L2Gateway."""
        gw = l2_gateway[self.gateway_resource]
        devices = gw.get('devices')
        dev_db = None
        l2gw_db = None
        with db_api.CONTEXT_WRITER.using(context):
            l2gw_db = self._get_l2_gateway(context, id)
            if not devices and l2gw_db:
                l2gw_db.name = gw.get('name')
                return self._make_l2_gateway_dict(l2gw_db)
            if devices:
                for device in devices:
                    dev_name = device['device_name']
                    dev_db = (self._get_l2gw_devices_by_name_andl2gwid(
                        context, dev_name, id))
                    interface_dict_list = [i for i in device['interfaces']]
                    interface_db = self._get_l2_gw_interfaces(context,
                                                              dev_db[0].id)
                    self._delete_l2_gateway_interfaces(context, interface_db)
                    self._update_interfaces_db(context, interface_dict_list,
                                               dev_db)
        if l2gw_db:
            if gw.get('name'):
                l2gw_db.name = gw.get('name')
            return self._make_l2_gateway_dict(l2gw_db)

    def _update_interfaces_db(self, context, interface_dict_list, device_db):
        for interfaces in interface_dict_list:
            int_name = interfaces.get('name')
            if constants.SEG_ID in interfaces:
                seg_id_list = interfaces.get(constants.SEG_ID)
                for seg_ids in seg_id_list:
                    uuid = self._generate_uuid()
                    int_db = self._get_int_model(uuid,
                                                 int_name,
                                                 device_db[0].id,
                                                 seg_ids)
                    context.session.add(int_db)
            else:
                uuid = self._generate_uuid()
                interface_db = self._get_int_model(uuid,
                                                   int_name,
                                                   device_db[0].id,
                                                   0)
                context.session.add(interface_db)

    @db_api.CONTEXT_READER
    def get_l2_gateway(self, context, id, fields=None):
        """get the l2 gateway by id."""
        self._admin_check(context, 'GET')
        gw_db = self._get_l2_gateway(context, id)
        return self._make_l2_gateway_dict(gw_db, fields)

    def delete_l2_gateway(self, context, id):
        """delete the l2 gateway  by id."""
        gw_db = self._get_l2_gateway(context, id)
        if gw_db:
            with db_api.CONTEXT_WRITER.using(context):
                context.session.delete(gw_db)
            LOG.debug("l2 gateway '%s' was deleted.", id)

    @db_api.CONTEXT_READER
    def get_l2_gateways(self, context, filters=None, fields=None,
                        sorts=None,
                        limit=None,
                        marker=None,
                        page_reverse=False):
        """list the l2 gateways available in the neutron DB."""
        self._admin_check(context, 'GET')
        marker_obj = self._get_marker_obj(
            context, 'l2_gateway', limit, marker)
        return self._get_collection(context, models.L2Gateway,
                                    self._make_l2_gateway_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    def _update_segmentation_id(self, context, l2gw_id, segmentation_id):
        """Update segmentation id for interfaces."""
        device_db = self._get_l2_gateway_devices(context, l2gw_id)
        for device_model in device_db:
            interface_db = self._get_l2_gw_interfaces(context,
                                                      device_model.id)
            for interface_model in interface_db:
                interface_model.segmentation_id = segmentation_id

    def _delete_l2_gateway_interfaces(self, context, int_db_list):
        """delete the l2 interfaces  by id."""
        with db_api.CONTEXT_WRITER.using(context):
            for interfaces in int_db_list:
                context.session.delete(interfaces)
        LOG.debug("l2 gateway interfaces was deleted.")

    def create_l2_gateway_connection(self, context, l2_gateway_connection):
        """Create l2 gateway connection."""
        gw_connection = l2_gateway_connection[self.connection_resource]
        l2_gw_id = gw_connection.get('l2_gateway_id')
        network_id = gw_connection.get('network_id')
        nw_map = {}
        nw_map['network_id'] = network_id
        nw_map['l2_gateway_id'] = l2_gw_id
        segmentation_id = ""
        if constants.SEG_ID in gw_connection:
            segmentation_id = gw_connection.get(constants.SEG_ID)
            nw_map[constants.SEG_ID] = segmentation_id
        with db_api.CONTEXT_WRITER.using(context):
            gw_db = self._get_l2_gateway(context, l2_gw_id)
            tenant_id = self._get_tenant_id_for_create(context, gw_db)
            nw_map['tenant_id'] = tenant_id
            if not segmentation_id:
                nw_map['segmentation_id'] = "0"
            conn_db = self._retrieve_gateway_connections(context, l2_gw_id,
                                                         nw_map)
            if not conn_db:
                connection_id = uuidutils.generate_uuid()
                nw_map['id'] = connection_id
                gw_db.network_connections.append(
                    models.L2GatewayConnection(**nw_map))
                gw_db = models.L2GatewayConnection(
                    id=connection_id,
                    tenant_id=tenant_id,
                    network_id=network_id,
                    l2_gateway_id=l2_gw_id,
                    segmentation_id=segmentation_id)
                return self._make_l2gw_connections_dict(gw_db)
            return self._make_l2gw_connections_dict(conn_db[0])

    @db_api.CONTEXT_READER
    def get_l2_gateway_connections_count(self, context, filters=None):
        return len(self._get_collection(context, models.L2GatewayConnection,
                                        self._make_l2gw_connections_dict,
                                        filters=filters))

    @db_api.CONTEXT_READER
    def get_l2_gateway_connections(self, context, filters=None,
                                   fields=None,
                                   sorts=None, limit=None, marker=None,
                                   page_reverse=False):
        """List l2 gateway connections."""
        self._admin_check(context, 'GET')
        marker_obj = self._get_marker_obj(
            context, 'l2_gateway_connection', limit, marker)
        return self._get_collection(context, models.L2GatewayConnection,
                                    self._make_l2gw_connections_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    @db_api.CONTEXT_READER
    def get_l2_gateway_connection(self, context, id, fields=None):
        """Get l2 gateway connection."""
        self._admin_check(context, 'GET')
        """Get the l2 gateway  connection  by id."""
        gw_db = self._get_l2_gateway_connection(context, id)
        return self._make_l2gw_connections_dict(gw_db, fields)

    def delete_l2_gateway_connection(self, context, id):
        """Delete the l2 gateway connection by id."""
        with db_api.CONTEXT_WRITER.using(context):
            gw_db = self._get_l2_gateway_connection(context, id)
            context.session.delete(gw_db)
        LOG.debug("l2 gateway '%s' was destroyed.", id)

    def _admin_check(self, context, action):
        """Admin role check helper."""
        # TODO(selva): his check should be required if the tenant_id is
        # specified in the request, otherwise the policy.json do a trick
        # this need further revision.
        if not context.is_admin:
            reason = _('Cannot %s resource for non admin tenant') % action
            raise exceptions.AdminRequired(reason=reason)

    def _generate_uuid(self):
        """Generate uuid helper."""
        uuid = uuidutils.generate_uuid()
        return uuid

    def _get_int_model(self, uuid, interface_name, dev_id, seg_id):
        return models.L2GatewayInterface(id=uuid,
                                         interface_name=interface_name,
                                         device_id=dev_id,
                                         segmentation_id=seg_id)

    def get_l2gateway_devices_by_gateway_id(self, context, l2_gateway_id):
        """Get l2gateway_devices_by id."""
        session = context.session
        with session.begin():
            return session.query(models.L2GatewayDevice).filter_by(
                l2_gateway_id=l2_gateway_id).all()

    def get_l2gateway_interfaces_by_device_id(self, context, device_id):
        """Get all l2gateway_interfaces_by device_id."""
        session = context.session
        with session.begin():
            return session.query(models.L2GatewayInterface).filter_by(
                device_id=device_id).all()

    def validate_device_name(self, context, device_name, l2gw_id):
        if device_name:
            devices_db = self._get_l2gw_devices_by_name_andl2gwid(context,
                                                                  device_name,
                                                                  l2gw_id)
        if not devices_db:
            raise l2gw_exc.L2GatewayDeviceNameNotFound(device_name=device_name)

    def _validate_any_seg_id_empty_in_interface_dict(self, devices):
        """Validate segmentation_id for consistency."""
        for device in devices:
            interface_list = device['interfaces']
            if not interface_list:
                raise l2gw_exc.L2GatewayInterfaceRequired()
            if constants.SEG_ID in interface_list[0]:
                for interfaces in interface_list[1:len(interface_list)]:
                    if constants.SEG_ID not in interfaces:
                        raise l2gw_exc.L2GatewaySegmentationRequired()
            if constants.SEG_ID not in interface_list[0]:
                for interfaces in interface_list[1:len(interface_list)]:
                    if constants.SEG_ID in interfaces:
                        raise l2gw_exc.L2GatewaySegmentationRequired()

    def _align_interfaces_list(self, interface_list):
        """Align interfaces list based on input dict for multiple seg ids."""
        interface_dict = {}
        aligned_interface_list = []
        for interfaces in interface_list:
            actual__name = interfaces.get('name')
            if actual__name in interface_dict:
                interface_name = interface_dict.get(actual__name)
                seg_id_list = interfaces.get(constants.SEG_ID)
                interface_name.append(str(seg_id_list))
                interface_dict.update({actual__name: interface_name})
            else:
                seg_id = str(interfaces.get(constants.SEG_ID)).split()
                interface_dict.update({actual__name: seg_id})
        for name in interface_dict:
            aligned_interface_list.append({'segmentation_id':
                                           interface_dict[name],
                                           'name': name})
        return aligned_interface_list

    @db_api.CONTEXT_READER
    def _get_l2_gateway_connections(self, context):
        """Get l2 gateway connections."""
        try:
            con = context.session.query(models.L2GatewayConnection).all()
        except sa_orm_exc.NoResultFound:
            raise l2gw_exc.L2GatewayConnectionNotFound(
                id="")
        return con

    @db_api.CONTEXT_READER
    def _get_l2gw_ids_by_interface_switch(self, context, interface_name,
                                          switch_name):
        """Get l2 gateway ids by interface and switch."""
        connections = self._get_l2_gateway_connections(context)
        l2gw_id_list = []
        if connections:
            for connection in connections:
                l2gw_id = connection.l2_gateway_id
                devices = self._get_l2_gateway_device_by_name_id(context,
                                                                 switch_name,
                                                                 l2gw_id)
                if devices:
                    for device in devices:
                        interfaces = self._get_l2_gw_interfaces(context,
                                                                device.id)
                        for interface in interfaces:
                            if interface_name == interface.interface_name:
                                l2gw_id_list.append(l2gw_id)
                else:
                    LOG.debug("l2 gateway devices are empty")
        else:
            LOG.debug("l2 gateway connections are empty")
        return l2gw_id_list

    def _delete_connection_by_l2gw_id(self, context, l2gw_id):
        """Delete the l2 gateway connection by l2gw id."""
        with db_api.CONTEXT_WRITER.using(context):
            con_db = self._get_l2_gateway_connection_by_l2gw_id(context,
                                                                l2gw_id)
            if con_db:
                context.session.delete(con_db[0])
                LOG.debug("l2 gateway connection was destroyed.")

    def _get_l2_gateway_connection_by_l2gw_id(self, context, l2gw_id):
        """Get the l2 gateway connection by l2gw id."""
        try:
            con = context.session.query(models.L2GatewayConnection).filter_by(
                l2_gateway_id=l2gw_id).all()
        except sa_orm_exc.NoResultFound:
            raise l2gw_exc.L2GatewayConnectionNotFound(
                id=l2gw_id)
        return con

    def _get_l2_gateway_device_by_name_id(self, context, device_name, l2gw_id):
        """Get the l2 gateway device by name and id."""
        try:
            gw = context.session.query(models.L2GatewayDevice).filter_by(
                device_name=device_name, l2_gateway_id=l2gw_id).all()
        except sa_orm_exc.NoResultFound:
            raise l2gw_exc.L2GatewayDeviceNotFound(
                device_id=device_name)
        return gw

    def validate_l2_gateway_for_create(self, context, l2_gateway):
        self._admin_check(context, 'CREATE')
        gw = l2_gateway[self.gateway_resource]
        devices = gw['devices']
        self._validate_any_seg_id_empty_in_interface_dict(devices)

    def validate_l2_gateway_for_delete(self, context, l2gw_id):
        self._admin_check(context, 'DELETE')
        gw_db = self._get_l2_gateway(context, l2gw_id)
        if gw_db.network_connections:
            raise l2gw_exc.L2GatewayInUse(gateway_id=l2gw_id)
        return

    def validate_l2_gateway_for_update(self, context, id, l2_gateway):
        self._admin_check(context, 'UPDATE')
        gw = l2_gateway[self.gateway_resource]
        devices = None
        dev_db = None
        if 'devices' in gw:
            devices = gw['devices']
        if not devices:
            return
        with db_api.CONTEXT_READER.using(context):
            # Attemp to retrieve l2gw
            gw_db = self._get_l2_gateway(context, id)
            if devices:
                for device in devices:
                    dev_name = device['device_name']
                    dev_db = self._get_l2gw_devices_by_name_andl2gwid(
                        context, dev_name, id)
                    if not dev_db:
                        raise l2gw_exc.L2GatewayDeviceNotFound(
                            device_id=dev_name)
                    self.validate_device_name(context, dev_name, id)
                    interface_list = device['interfaces']
                    if not interface_list:
                        raise l2gw_exc.L2GatewayInterfaceRequired()
                    if constants.SEG_ID in interface_list[0]:
                        for interfaces in interface_list:
                            if constants.SEG_ID not in interfaces:
                                raise l2gw_exc.L2GatewaySegmentationRequired()
                    if constants.SEG_ID not in interface_list[0]:
                        for interfaces in interface_list[1:]:
                            if constants.SEG_ID in interfaces:
                                raise l2gw_exc.L2GatewaySegmentationRequired()
                    if gw_db.devices:
                        if gw_db.devices[0].interfaces[0]['segmentation_id']:
                            if constants.SEG_ID not in interface_list[0]:
                                raise l2gw_exc.L2GatewaySegmentationIDExists()
                        else:
                            if constants.SEG_ID in interface_list[0]:
                                raise l2gw_exc.\
                                    L2GatewaySegmentationIDNotExists()

    def validate_l2_gateway_connection_for_create(self, context,
                                                  l2_gateway_connection):
        self._admin_check(context, 'CREATE')
        gw_connection = l2_gateway_connection[self.connection_resource]
        l2_gw_id = gw_connection.get('l2_gateway_id')
        network_id = gw_connection.get('network_id')
        plugin = directory.get_plugin()
        plugin.get_network(context, network_id)
        nw_map = {}
        nw_map['network_id'] = network_id
        nw_map['l2_gateway_id'] = l2_gw_id
        segmentation_id = ""
        if constants.SEG_ID in gw_connection:
            segmentation_id = gw_connection.get(constants.SEG_ID)
            nw_map[constants.SEG_ID] = segmentation_id
        is_vlan = self._is_vlan_configured_on_any_interface_for_l2gw(context,
                                                                     l2_gw_id)
        network_id = l2gw_validators.validate_network_mapping_list(nw_map,
                                                                   is_vlan)
        with db_api.CONTEXT_WRITER.using(context):
            if self._retrieve_gateway_connections(context,
                                                  l2_gw_id,
                                                  nw_map):
                raise l2gw_exc.L2GatewayConnectionExists(mapping=nw_map,
                                                         gateway_id=l2_gw_id)

    @db_api.CONTEXT_READER
    def validate_l2_gateway_connection_for_delete(self, context,
                                                  l2_gateway_conn_id):
        self._admin_check(context, 'DELETE')
        gw_db = self._get_l2_gateway_connection(context,
                                                l2_gateway_conn_id)
        if gw_db is None:
            raise l2gw_exc.L2GatewayConnectionNotFound(
                id=l2_gateway_conn_id)


def l2gw_callback(resource, event, trigger, **kwargs):
    l2gwservice = directory.get_plugin(constants.L2GW)
    context = kwargs.get('context')
    port_dict = kwargs.get('port')
    if l2gwservice:
        if event == events.AFTER_UPDATE:
            l2gwservice.add_port_mac(context, port_dict)
        elif event == events.AFTER_DELETE:
            l2gwservice.delete_port_mac(context, port_dict)


def subscribe():
    interested_events = (events.AFTER_UPDATE,
                         events.AFTER_DELETE)
    for x in interested_events:
        registry.subscribe(
            l2gw_callback, resources.PORT, x)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import directory
from oslo_utils import excutils

from neutron.db import servicetype_db as st_db
from neutron.services import provider_configuration as pconf
from neutron.services import service_base

from networking_l2gw.db.l2gateway import l2gateway_db
from networking_l2gw.services.l2gateway.common import config
from networking_l2gw.services.l2gateway.common import constants
from networking_l2gw.services.l2gateway import exceptions as exc

from neutron_lib.db import api as db_api
from neutron_lib import exceptions as n_exc
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class L2GatewayPlugin(l2gateway_db.L2GatewayMixin):

    """Implementation of the Neutron l2 gateway Service Plugin.

    This class manages the workflow of L2 gateway request/response.
    """

    supported_extension_aliases = ["l2-gateway",
                                   "l2-gateway-connection"]

    def __init__(self):
        """Do the initialization for the l2 gateway service plugin here."""
        config.register_l2gw_opts_helper()
        self.service_type_manager = st_db.ServiceTypeManager.get_instance()
        self.service_type_manager.add_provider_configuration(
            constants.L2GW, pconf.ProviderConfiguration('networking_l2gw'))
        self._load_drivers()
        LOG.info("L2Gateway Service Plugin using Service Driver: %s",
                 self.default_provider)
        self.driver = self.drivers[self.default_provider]
        if len(self.drivers) > 1:
            LOG.warning("Multiple drivers configured for L2Gateway, "
                        "although running multiple drivers in parallel"
                        " is not yet supported")
        super(L2GatewayPlugin, self).__init__()
        l2gateway_db.subscribe()

    def _load_drivers(self):
        """Loads plugin-drivers specified in configuration."""
        self.drivers, self.default_provider = service_base.load_drivers(
            'L2GW', self)

    def _get_driver_for_provider(self, provider):
        if provider in self.drivers:
            return self.drivers[provider]
        # raise if not associated (should never be reached)
        raise n_exc.Invalid("Error retrieving driver for provider %s" %
                            provider)

    @property
    def _core_plugin(self):
        return directory.get_plugin()

    def get_plugin_type(self):
        """Get type of the plugin."""
        return constants.L2GW

    def get_plugin_description(self):
        """Get description of the plugin."""
        return constants.L2_GATEWAY_SERVICE_PLUGIN

    def add_port_mac(self, context, port_dict):
        """Process a created Neutron port."""
        self.driver.add_port_mac(context, port_dict)

    def delete_port_mac(self, context, port):
        """Process a deleted Neutron port."""
        self.driver.delete_port_mac(context, port)

    def create_l2_gateway(self, context, l2_gateway):
        """Create the L2Gateway."""
        self.validate_l2_gateway_for_create(context, l2_gateway)
        self.driver.create_l2_gateway(context, l2_gateway)
        with db_api.CONTEXT_WRITER.using(context):
            l2_gateway_instance = super(L2GatewayPlugin,
                                        self).create_l2_gateway(context,
                                                                l2_gateway)
            self.driver.create_l2_gateway_precommit(context,
                                                    l2_gateway_instance)
        try:
            self.driver.create_l2_gateway_postcommit(
                context, l2_gateway_instance)
        except exc.L2GatewayServiceDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error("L2GatewayPlugin.create_l2_gateway_postcommit "
                          "failed, deleting l2gateway '%s'",
                          l2_gateway_instance['id'])
                self.delete_l2_gateway(context, l2_gateway_instance['id'])
        return l2_gateway_instance

    def update_l2_gateway(self, context, l2_gateway_id, l2_gateway):
        """Update the L2Gateway."""
        self.validate_l2_gateway_for_update(context, l2_gateway_id, l2_gateway)
        self.driver.update_l2_gateway(context, l2_gateway_id, l2_gateway)
        with db_api.CONTEXT_WRITER.using(context):
            l2_gateway_instance = super(L2GatewayPlugin,
                                        self).update_l2_gateway(context,
                                                                l2_gateway_id,
                                                                l2_gateway)
            self.driver.update_l2_gateway_precommit(context,
                                                    l2_gateway_instance)
        try:
            self.driver.update_l2_gateway_postcommit(
                context, l2_gateway_instance)
        except exc.L2GatewayServiceDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error("L2GatewayPlugin.update_l2_gateway_postcommit"
                          " failed for l2gateway %s", l2_gateway_id)
        return l2_gateway_instance

    def delete_l2_gateway(self, context, l2_gateway_id):
        """Delete the L2Gateway."""
        self.validate_l2_gateway_for_delete(context, l2_gateway_id)
        self.driver.delete_l2_gateway(context, l2_gateway_id)
        with db_api.CONTEXT_WRITER.using(context):
            super(L2GatewayPlugin, self).delete_l2_gateway(context,
                                                           l2_gateway_id)
            self.driver.delete_l2_gateway_precommit(context, l2_gateway_id)
        try:
            self.driver.delete_l2_gateway_postcommit(context, l2_gateway_id)
        except exc.L2GatewayServiceDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error("L2GatewayPlugin.delete_l2_gateway_postcommit"
                          " failed for l2gateway %s", l2_gateway_id)

    def create_l2_gateway_connection(self, context, l2_gateway_connection):
        """Create an L2Gateway Connection

        Bind a Neutron VXLAN Network to Physical Network Segment.
        """
        self.validate_l2_gateway_connection_for_create(
            context, l2_gateway_connection)
        self.driver.create_l2_gateway_connection(context,
                                                 l2_gateway_connection)
        with db_api.CONTEXT_WRITER.using(context):
            l2_gateway_conn_instance = super(
                L2GatewayPlugin, self).create_l2_gateway_connection(
                    context, l2_gateway_connection)
            self.driver.create_l2_gateway_connection_precommit(
                context, l2_gateway_conn_instance)
        try:
            self.driver.create_l2_gateway_connection_postcommit(
                context, l2_gateway_conn_instance)
        except exc.L2GatewayServiceDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error("L2GatewayPlugin."
                          "create_l2_gateway_connection_postcommit "
                          "failed, deleting connection '%s'",
                          l2_gateway_conn_instance['id'])
                self.delete_l2_gateway_connection(
                    context, l2_gateway_conn_instance['id'])

        return l2_gateway_conn_instance

    def delete_l2_gateway_connection(self, context, l2_gateway_connection_id):
        """Delete an L2Gateway Connection

        Unbind a Neutron VXLAN Network from Physical Network Segment.
        """
        self.validate_l2_gateway_connection_for_delete(
            context, l2_gateway_connection_id)
        self.driver.delete_l2_gateway_connection(
            context, l2_gateway_connection_id)
        with db_api.CONTEXT_WRITER.using(context):
            super(L2GatewayPlugin, self).delete_l2_gateway_connection(
                context, l2_gateway_connection_id)
            self.driver.delete_l2_gateway_connection_precommit(
                context, l2_gateway_connection_id)
        try:
            self.driver.delete_l2_gateway_connection_postcommit(
                context, l2_gateway_connection_id)
        except exc.L2GatewayServiceDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(
                    "L2GatewayPlugin.delete_l2_gateway_connection_postcommit"
                    " failed for connection %s", l2_gateway_connection_id)



# Copyright 2017 <PUT YOUR NAME/COMPANY HERE>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add standard_attribute_id for l2gw

Revision ID: 0fb45e525aa9
Revises: 2f533f7705dd
Create Date: 2017-12-19 07:49:35.418145

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '0fb45e525aa9'
down_revision = '2f533f7705dd'
depends_on = ('8d7d772eafcf',)

TABLES = ("l2gateways", "l2gatewayconnections", "l2gatewaydevices")

TABLE_MODELS = [
    (table, sa.Table(table, sa.MetaData(),
                     sa.Column('id', sa.String(length=36), nullable=False),
                     sa.Column('standard_attr_id', sa.BigInteger(),
                               nullable=True)))
    for table in TABLES
]

standardattrs = sa.Table(
    'standardattributes', sa.MetaData(),
    sa.Column('id', sa.BigInteger(), primary_key=True, autoincrement=True),
    sa.Column('resource_type', sa.String(length=255), nullable=False))


def upgrade():
    generate_records_for_existing()
    for table, model in TABLE_MODELS:
        op.alter_column(table, 'standard_attr_id', nullable=False,
                        existing_type=sa.BigInteger(), existing_nullable=True,
                        existing_server_default=False)
        op.create_foreign_key(
            constraint_name=None, source_table=table,
            referent_table='standardattributes',
            local_cols=['standard_attr_id'], remote_cols=['id'],
            ondelete='CASCADE')
        op.create_unique_constraint(
            constraint_name='uniq_%s0standard_attr_id' % table,
            table_name=table, columns=['standard_attr_id'])


def generate_records_for_existing():
    session = sa.orm.Session(bind=op.get_bind())
    values = []
    for table, model in TABLE_MODELS:
        for row in session.query(model):
            res = session.execute(
                standardattrs.insert().values(resource_type=table))
            session.execute(
                model.update().values(
                    standard_attr_id=res.inserted_primary_key[0]).where(
                        model.c.id == row[0]))
    # this commit is necessary to allow further operations
    session.commit()
    return values



# Copyright 2015 OpenStack Foundation
# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from neutron.tests.unit import testlib_api
from neutron_lib.callbacks import events
from neutron_lib.callbacks import resources
from neutron_lib import context
from neutron_lib.db import api as db_api

from networking_l2gw.db.l2gateway import l2gateway_db
from networking_l2gw.services.l2gateway.common import constants
from networking_l2gw.services.l2gateway.common import l2gw_validators
from networking_l2gw.services.l2gateway import exceptions

from neutron_lib import exceptions as exc
from neutron_lib.plugins import directory
from oslo_utils import importutils
from oslo_utils import uuidutils

DB_PLUGIN_KLASS = 'neutron.db.db_base_plugin_v2.NeutronDbPluginV2'
_uuid = uuidutils.generate_uuid


class L2GWTestCase(testlib_api.SqlTestCase):

    """Unit test for l2 Gateway DB support."""

    def setUp(self):
        super(L2GWTestCase, self).setUp()
        self.ctx = context.get_admin_context()
        self.mixin = l2gateway_db.L2GatewayMixin()
        self.gw_resource = constants.L2_GATEWAYS
        self.con_resource = constants.CONNECTION_RESOURCE_NAME
        self.plugin = importutils.import_object(DB_PLUGIN_KLASS)

    def _create_l2gateway(self, l2gateway):
        """Create l2gateway helper method."""
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self.mixin.create_l2_gateway(self.ctx, l2gateway)

    def _get_l2_gateway_data(self, name, device_name):
        """Get l2 gateway data helper method."""
        data = {"l2_gateway": {"name": name,
                               "devices":
                               [{"interfaces": [{"name": "port1",
                                                 "segmentation_id": ["111"]}],
                                 "device_name": device_name}]}}
        return data

    def _get_l2_gateway_data_with_multiple_segid(self, name, device_name):
        """Get l2 gateway data helper method for multiple seg id."""
        data = {"l2_gateway": {"name": name,
                               "devices":
                               [{"interfaces": [{"name": "port1",
                                                 "segmentation_id": ["111",
                                                                     "123"]}],
                                 "device_name": device_name}]}}
        return data

    def _get_l2_gateway_multiple_interface_data(self, name, device_name):
        """Get l2 gateway data helper method with multiple interface data."""
        data = {"l2_gateway": {"name": name,
                               "devices":
                               [{"interfaces": [{"name": "port1",
                                                 "segmentation_id": ["4076"]},
                                                {"name": "port1",
                                                 "segmentation_id": ["4074"]}],
                                 "device_name": device_name}]}}
        return data

    def _get_l2_gw_multiple_interface_partial_seg_id_data(self, name,
                                                          device_name):
        """Get l2 gateway data helper method with partial seg id."""
        data = {"l2_gateway": {"name": name,
                               "devices":
                               [{"interfaces": [{"name": "port1",
                                                 "segmentation_id": ["4076"]},
                                                {"name": "port1"}],
                                 "device_name": device_name}]}}
        return data

    def _get_l2_gw_multiple_interface_without_seg_id_data(self, name,
                                                          device_name):
        """Get l2 gateway data helper method with partial seg id."""
        return {"l2_gateway": {"name": name,
                               "devices":
                               [{"interfaces": [{"name": "port1"},
                                                {"name": "port2"}],
                                 "device_name": device_name}]}}

    def _get_l2_gw_invalid_seg_id_data(self, name,
                                       device_name):
        """Get l2 gateway data helper method with invalid seg id."""
        data = {"interfaces": [{"name": "port1",
                                "segmentation_id": ["test"]}],
                "device_name": device_name}
        return [data]

    def _get_nw_data(self):
        return {'network': {'id': _uuid(),
                            'name': 'net1',
                            'admin_state_up': True,
                            'tenant_id': 'test-tenant',
                            'shared': False}}

    def _get_l2_gateway_data_without_seg_id(self, name, device_name):
        """Get l2 gateway data helper method."""
        data = {"l2_gateway": {"name": name,
                               "devices":
                               [{"interfaces": [{"name": "port1"}],
                                "device_name": device_name}]}}
        return data

    def test_l2_gateway_get(self):
        """Test l2 gateway get."""
        name = "l2gw_1"
        device_name = "device1"
        data = self._get_l2_gateway_data(name, device_name)
        result = self._create_l2gateway(data)
        get_result = self._get_l2_gateway(result['id'])
        self.assertEqual(name, get_result['name'])

    def test_l2_gateway_get_invalid_id_failure(self):
        """Test l2 gateway get for an invalid L2 gateway UUID."""
        # Generate a random UUID and try to retrieve a L2 gateway
        # using that UUID.
        self.assertRaises(exceptions.L2GatewayNotFound,
                          self._get_l2_gateway, _uuid())

    def test_l2_gateway_create(self):
        """Test l2 gateway create."""
        name = "l2gw_1"
        device_name = "device1"
        data = self._get_l2_gateway_data(name, device_name)
        result = self._create_l2gateway(data)
        self.assertEqual(result['name'], name)

    def _get_l2_gateway(self, l2gw_id):
        with db_api.CONTEXT_READER.using(self.ctx):
            return self.mixin.get_l2_gateway(self.ctx, l2gw_id)

    def _get_l2_gateways(self):
        """Update l2gateway helper."""
        with db_api.CONTEXT_READER.using(self.ctx):
            return self.mixin.get_l2_gateways(self.ctx)

    def test_l2gateway_list(self):
        """Test l2 gateway list."""
        name = "l2gw_1"
        device_name = "device1"
        data = self._get_l2_gateway_data(name, device_name)
        self._create_l2gateway(data)
        result2 = self._get_l2_gateways()
        self.assertIn('id', result2[0])

    def test_l2gateway_show(self):
        """Test l2 gateway show."""
        name = "l2gw_1"
        device_name = "device1"
        data = self._get_l2_gateway_data(name, device_name)
        gw = self._create_l2gateway(data)
        l2gw_id = gw['id']
        result = self._get_l2_gateway(l2gw_id)
        self.assertEqual(name, result['name'])

    def _update_l2_gateway(self, id, l2gateway):
        """Update l2gateway helper."""
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self.mixin.update_l2_gateway(self.ctx, id, l2gateway)

    def test_l2_gateway_update(self):
        """Test l2 gateway update."""
        name_create = "l2gw_1"
        name_update = "l2gw_2"
        device_name = "device1"
        data_l2gw_create = self._get_l2_gateway_data(name_create,
                                                     device_name)
        gw_org = self._create_l2gateway(data_l2gw_create)
        l2gw_id = gw_org['id']
        l2_gw_update_dict = self._get_l2_gateway_data(name_update,
                                                      device_name)
        result = self._update_l2_gateway(l2gw_id, l2_gw_update_dict)
        self.assertNotEqual(result['name'], name_create)

    def test_l2_gateway_update_without_devices(self):
        """Test l2 gateway update without devices."""
        name_create = "l2gw_1"
        name_update = "l2gw_updated"
        device_name = "device1"
        data_l2gw_create = self._get_l2_gateway_data(name_create,
                                                     device_name)
        gw_org = self._create_l2gateway(data_l2gw_create)
        l2gw_id = gw_org['id']
        l2_gw_update_dict = {"l2_gateway": {"name": name_update}}
        result = self._update_l2_gateway(l2gw_id, l2_gw_update_dict)
        self.assertNotEqual(result['name'], name_create)
        self.assertEqual(result['name'], name_update)

    def _create_l2gateway_connection(self, l2gateway_con):
        """Create L2 gateway connection resource helper method."""
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self.mixin.create_l2_gateway_connection(self.ctx,
                                                           l2gateway_con)

    def _list_l2gateway_connection(self):
        """Create L2 gateway connection resource helper method."""
        with db_api.CONTEXT_READER.using(self.ctx):
            return self.mixin.get_l2_gateway_connections(self.ctx)

    def _get_no_l2gateway_connections(self):
        """Create L2 gateway connection resource helper method."""
        with db_api.CONTEXT_READER.using(self.ctx):
            return self.mixin.get_l2_gateway_connections_count(self.ctx)

    def _delete_l2gw_connection_by_l2gw_id(self, l2gw_id):
        """Delete l2 gateway connection."""
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self.mixin._delete_connection_by_l2gw_id(self.ctx, l2gw_id)

    def test_get_l2gw_ids_by_interface_switch(self):
        """Test get L2 gateway ids by interface and switch name."""
        name = "l2gw_con1"
        device_name = "device1"
        data_l2gw = self._get_l2_gateway_data(name, device_name)
        gw = self._create_l2gateway(data_l2gw)
        net_data = self._get_nw_data()
        net = self.plugin.create_network(self.ctx, net_data)
        l2gw_id = gw['id']
        data_con = {self.con_resource: {'l2_gateway_id': l2gw_id,
                                        'network_id': net['id']}}

        self._create_l2gateway_connection(data_con)
        l2gw_id_list = self.mixin._get_l2gw_ids_by_interface_switch(
            self.ctx, 'port1', 'device1')
        self.assertEqual(l2gw_id_list[0], l2gw_id)

    def test_l2gateway_connection_create_delete_list(self):
        """Test l2 gateway connection create and delete."""
        name = "l2gw_con1"
        device_name = "device_name1"
        data_l2gw = self._get_l2_gateway_data(name, device_name)
        gw = self._create_l2gateway(data_l2gw)
        net_data = self._get_nw_data()
        net = self.plugin.create_network(self.ctx, net_data)
        l2gw_id = gw['id']
        data_con = {self.con_resource: {'l2_gateway_id': l2gw_id,
                                        'network_id': net['id']}}
        gw_con = self._create_l2gateway_connection(data_con)
        exp_net_id = gw_con['network_id']
        self.assertEqual(net['id'], exp_net_id)
        list_con = self._list_l2gateway_connection()
        self.assertEqual(1, self._get_no_l2gateway_connections())
        self.assertIn('id', list_con[0])
        result = self._delete_l2gw_connection_by_l2gw_id(l2gw_id)
        self.assertIsNone(result)

    def _validate_l2_gateway_for_delete(self, l2gw_id):
        """Delete l2 gateway helper method."""
        with db_api.CONTEXT_READER.using(self.ctx):
            return self.mixin.validate_l2_gateway_for_delete(self.ctx, l2gw_id)

    def _validate_l2_gateway_for_create(self, l2gw):
        """Create l2 gateway helper method."""
        with db_api.CONTEXT_READER.using(self.ctx):
            return self.mixin.validate_l2_gateway_for_create(self.ctx, l2gw)

    def _validate_l2_gateway_for_update(self, l2gw_id, l2gw):
        """Update l2 gateway helper method."""
        with db_api.CONTEXT_READER.using(self.ctx):
            return self.mixin.validate_l2_gateway_for_update(self.ctx,
                                                             l2gw_id, l2gw)

    def _validate_l2_gateway_connection_for_create(self, l2gw_con):
        """Create l2 gateway connection helper method."""
        with db_api.CONTEXT_READER.using(self.ctx):
            return self.mixin.validate_l2_gateway_connection_for_create(
                self.ctx, l2gw_con)

    def test_l2gateway_con_create_and_delete_in_use_without_seg_id(self):
        """Test l2 gateway connection create without seg id when use."""
        name = "l2gw_con2"
        device_name = "device_name2"
        data_l2gw = self._get_l2_gateway_data(name,
                                              device_name)
        gw = self._create_l2gateway(data_l2gw)
        net_data = self._get_nw_data()
        net = self.plugin.create_network(self.ctx, net_data)
        l2gw_id = gw['id']
        data_con = {self.con_resource: {'l2_gateway_id': l2gw_id,
                                        'network_id': net['id']}}
        self._create_l2gateway_connection(data_con)
        self.assertRaises(exceptions.L2GatewayInUse,
                          self._validate_l2_gateway_for_delete, l2gw_id)

    def test_l2gateway_con_create_with_invalid_net_id(self):
        """Test l2 gateway connection create with invalid net id."""
        name = "l2gw_con2"
        device_name = "device_name2"
        data_l2gw = self._get_l2_gateway_data(name,
                                              device_name)
        gw = self._create_l2gateway(data_l2gw)
        net_id = 'invalid_net_id'
        l2gw_id = gw['id']
        data_con = {self.con_resource: {'l2_gateway_id': l2gw_id,
                                        'network_id': net_id}}
        directory.add_plugin('CORE', self.plugin)
        self.assertRaises(exc.NetworkNotFound,
                          self._validate_l2_gateway_connection_for_create,
                          data_con)

    def _delete_l2gateway(self, l2gw_id):
        """Delete l2 gateway helper method."""
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self.mixin.delete_l2_gateway(self.ctx,
                                                l2gw_id)

    def test_l2gateway_delete(self):
        """Test l2 gateway delete."""
        data_l2gw = self._get_l2_gateway_data("gateway_delete",
                                              "device_name")
        gw_actual = self._create_l2gateway(data_l2gw)
        l2gw_id = gw_actual['id']
        result = self._delete_l2gateway(l2gw_id)
        self.assertIsNone(result)

    def _delete_l2gw_connection(self, con_id):
        """Delete l2 gateway connection."""
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self.mixin.delete_l2_gateway_connection(self.ctx, con_id)

    def test_l2_gateway_create_with_mul_interfaces(self):
        """Test l2 gateway create with multiple interfaces all seg id."""
        name = "l2gw_1"
        device_name = "device1"
        data = self._get_l2_gateway_multiple_interface_data(name, device_name)
        result = self._create_l2gateway(data)
        self.assertEqual(result['name'], name)

    def test_l2_gateway_create_with_mul_interfaces_inconsistent_seg_id(self):
        """Test l2 gateway create with multiple interfaces."""
        name = "l2gw_1"
        dev_name = "device1"
        data = self._get_l2_gw_multiple_interface_partial_seg_id_data(name,
                                                                      dev_name)
        self.assertRaises(exceptions.L2GatewaySegmentationRequired,
                          self._validate_l2_gateway_for_create, data)

    def test_l2_gateway_update_with_mul_interfaces_inconsistent_seg_id(self):
        """Test l2 gateway update with multiple interfaces."""
        name = "l2gw_1"
        dev_name = "device1"
        data_orig = self._get_l2_gateway_multiple_interface_data(name,
                                                                 dev_name)
        data = self._get_l2_gw_multiple_interface_partial_seg_id_data(name,
                                                                      dev_name)
        gw_org = self._create_l2gateway(data_orig)
        l2gw_id = gw_org['id']
        self.assertRaises(exceptions.L2GatewaySegmentationRequired,
                          self._validate_l2_gateway_for_update, l2gw_id, data)

    def test_l2_gateway_update_with_mul_interfaces_without_seg_id(self):
        """Test l2 gateway update with multiple interfaces without seg_id."""
        name = "l2gw_1"
        dev_name = "device1"
        data1 = self._get_l2_gateway_multiple_interface_data(
            name, dev_name)
        data2 = self._get_l2_gw_multiple_interface_without_seg_id_data(
            name, dev_name)
        gw_org1 = self._create_l2gateway(data1)
        l2gw_id1 = gw_org1['id']
        gw_org2 = self._create_l2gateway(data2)
        l2gw_id2 = gw_org2['id']
        self.assertRaises(exceptions.L2GatewaySegmentationIDExists,
                          self._validate_l2_gateway_for_update,
                          l2gw_id1, data2)
        self.assertRaises(exceptions.L2GatewaySegmentationIDNotExists,
                          self._validate_l2_gateway_for_update,
                          l2gw_id2, data1)

    def test_l2_gateway_create_with_invalid_seg_id(self):
        """Test l2 gateway create with invalid seg-id."""
        name = "l2gw_1"
        dev_name = "device1"
        data = self._get_l2_gw_invalid_seg_id_data(name, dev_name)
        self.assertRaises(exc.InvalidInput,
                          l2gw_validators.validate_gwdevice_list, data)

    def test_l2_gateway_create_with_multiple_segid(self):
        """Test l2 gateway create with multiple seg id."""
        name = "l2gw_1"
        device_name = "device1"
        data = self._get_l2_gateway_data_with_multiple_segid(name, device_name)
        result = self._create_l2gateway(data)
        self.assertEqual(result['name'], name)

    def test_l2_gateway_update_invalid_device_name(self):
        """Test l2 gateway update with invalid device name."""
        name_create = "l2gw_1"
        device_name = "device1"
        invalid_device_name = "invalid_device"
        data_l2gw_create = self._get_l2_gateway_data(name_create,
                                                     device_name)
        data_l2gw_update = self._get_l2_gateway_data(name_create,
                                                     invalid_device_name)
        gw_org = self._create_l2gateway(data_l2gw_create)
        l2gw_id = gw_org['id']
        self.assertRaises(exceptions.L2GatewayDeviceNotFound,
                          self._validate_l2_gateway_for_update, l2gw_id,
                          data_l2gw_update)

    def test_l2gw_callback_update_port(self):
        service_plugin = mock.Mock()
        directory.add_plugin(constants.L2GW, service_plugin)
        fake_context = mock.Mock()
        fake_port = mock.Mock()
        fake_kwargs = {'context': fake_context,
                       'port': fake_port}
        l2gateway_db.l2gw_callback(resources.PORT,
                                   events.AFTER_UPDATE,
                                   mock.Mock(),
                                   **fake_kwargs)
        self.assertTrue(service_plugin.add_port_mac.called)

    def test_l2gw_callback_delete_port(self):
        service_plugin = mock.Mock()
        directory.add_plugin(constants.L2GW, service_plugin)
        fake_context = mock.Mock()
        fake_port = mock.Mock()
        fake_kwargs = {'context': fake_context,
                       'port': fake_port}
        l2gateway_db.l2gw_callback(resources.PORT,
                                   events.AFTER_DELETE,
                                   mock.Mock(),
                                   **fake_kwargs)
        self.assertTrue(service_plugin.delete_port_mac.called)

    def test_l2_gateway_create_output_aligned_with_input(self):
        """Test l2 gateway create output that is aligned with input dict."""
        name = "l2gw_1"
        device_name = "device1"
        data = self._get_l2_gateway_data_with_multiple_segid(name, device_name)
        result = self._create_l2gateway(data)
        gw_input = data['l2_gateway']
        devices_input = gw_input['devices']
        devices_output = result['devices']
        input_seg_list = devices_input[0]['interfaces'][0]['segmentation_id']
        output_seg_list = devices_output[0]['interfaces'][0]['segmentation_id']
        self.assertEqual(len(input_seg_list), len(output_seg_list))

    def test_l2gateway_show_update_delete_invalid_id(self):
        """Test l2 gateway show, update and delete with invalid id."""
        name = "l2gw_1"
        name_update = "l2gw_2"
        device_name = "device1"
        invalid_l2gw_id = "invalid_id"
        data_l2gw_create = self._get_l2_gateway_data(name, device_name)
        data_l2gw_update = self._get_l2_gateway_data(name_update, device_name)
        self._create_l2gateway(data_l2gw_create)
        self.assertRaises(exceptions.L2GatewayNotFound,
                          self._get_l2_gateway, invalid_l2gw_id)
        self.assertRaises(exceptions.L2GatewayNotFound,
                          self._validate_l2_gateway_for_update,
                          invalid_l2gw_id, data_l2gw_update)
        self.assertRaises(exceptions.L2GatewayNotFound,
                          self._validate_l2_gateway_for_delete,
                          invalid_l2gw_id)



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_db import exception as d_exc
from oslo_utils import timeutils
from oslo_utils import uuidutils

from neutron.tests.unit import testlib_api
from neutron_lib import context
from neutron_lib.db import api as db_api

from networking_l2gw.db.l2gateway.ovsdb import lib
from networking_l2gw.db.l2gateway.ovsdb import models

_uuid = uuidutils.generate_uuid


class OvsdbLibTestCase(testlib_api.SqlTestCase):

    def setUp(self):
        super(OvsdbLibTestCase, self).setUp()
        self.ctx = context.get_admin_context()

    def _get_logical_switch_dict(self):
        uuid = _uuid()
        record_dict = {'uuid': uuid,
                       'name': 'logical_switch1',
                       'key': '100',
                       'ovsdb_identifier': "host1"}
        return record_dict

    def _create_logical_switch(self, record_dict, name=None):
        if name:
            record_dict['name'] = name
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = models.LogicalSwitches(
                uuid=record_dict['uuid'],
                name=record_dict['name'],
                key=record_dict['key'],
                ovsdb_identifier=record_dict['ovsdb_identifier'])
            self.ctx.session.add(entry)
            return entry

    def _assert_model_equals(self, entry, result):
        for field, res in result.items():
            self.assertEqual(str(entry[field]), str(res))

    def test_get_logical_switch(self):
        record_dict = self._get_logical_switch_dict()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_logical_switch(record_dict)
        result = lib.get_logical_switch(self.ctx, record_dict)
        self._assert_model_equals(entry, result)

    def test_get_logical_switch_return_none(self):
        record_dict = {'uuid': 'foo_uuid', 'ovsdb_identifier': 'foo_ovsdb_id'}
        result = lib.get_logical_switch(self.ctx, record_dict)
        self.assertIsNone(result)

    def test_add_logical_switch(self):
        record_dict = self._get_logical_switch_dict()
        self._create_logical_switch(record_dict)
        count = self.ctx.session.query(models.LogicalSwitches).count()
        self.assertEqual(1, count)

    def test_delete_logical_switch(self):
        record_dict = self._get_logical_switch_dict()
        self._create_logical_switch(record_dict)
        lib.delete_logical_switch(self.ctx, record_dict)
        count = self.ctx.session.query(models.LogicalSwitches).count()
        self.assertEqual(count, 0)

    def _get_physical_locator_dict(self):
        uuid = _uuid()
        record_dict = {'uuid': uuid,
                       'dst_ip': '10.0.0.1',
                       'ovsdb_identifier': 'host1'}
        return record_dict

    def _create_physical_locator(self, record_dict, dst_ip=None):
        if dst_ip:
            record_dict['dst_ip'] = dst_ip
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = models.PhysicalLocators(
                uuid=record_dict['uuid'],
                dst_ip=record_dict['dst_ip'],
                ovsdb_identifier=record_dict['ovsdb_identifier'])
            self.ctx.session.add(entry)
            return entry

    def test_get_physical_locator(self):
        record_dict = self._get_physical_locator_dict()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_physical_locator(record_dict)
        result = lib.get_physical_locator(self.ctx, record_dict)
        self._assert_model_equals(entry, result)

    def test_add_physical_locator(self):
        record_dict = self._get_physical_locator_dict()
        self._create_physical_locator(record_dict)
        count = self.ctx.session.query(models.PhysicalLocators).count()
        self.assertEqual(1, count)

    def test_delete_physical_locator(self):
        record_dict = self._get_physical_locator_dict()
        self._create_physical_locator(record_dict)
        lib.delete_physical_locator(self.ctx, record_dict)
        count = self.ctx.session.query(models.PhysicalLocators).count()
        self.assertEqual(count, 0)

    def _get_physical_switch_dict(self):
        uuid = _uuid()
        record_dict = {'uuid': uuid,
                       'name': 'physical_switch1',
                       'tunnel_ip': '10.0.0.1',
                       'ovsdb_identifier': 'host1'}
        return record_dict

    def _create_physical_switch(self, record_dict, name=None):
        if name:
            record_dict['name'] = name
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = models.PhysicalSwitches(
                uuid=record_dict['uuid'],
                name=record_dict['name'],
                tunnel_ip=record_dict['tunnel_ip'],
                ovsdb_identifier=record_dict['ovsdb_identifier'])
            self.ctx.session.add(entry)
            return entry

    def test_get_physical_switch(self):
        record_dict = self._get_physical_switch_dict()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_physical_switch(record_dict)
        result = lib.get_physical_switch(self.ctx, record_dict)
        self._assert_model_equals(entry, result)

    def test_add_physical_switch(self):
        record_dict = self._get_physical_switch_dict()
        self._create_physical_switch(record_dict)
        count = self.ctx.session.query(models.PhysicalSwitches).count()
        self.assertEqual(1, count)

    def test_add_physical_switch_raise_on_duplicate_constraint(self):
        record_dict = self._get_physical_switch_dict()
        self._create_physical_switch(record_dict)
        # Call the method twice to trigger a db duplicate constraint error,
        # this time with a different switch name!
        self.assertRaises(d_exc.DBDuplicateEntry,
                          self._create_physical_switch,
                          record_dict, 'physical_switch2')

    def test_delete_physical_switch(self):
        record_dict = self._get_physical_switch_dict()
        self._create_physical_switch(record_dict)
        lib.delete_physical_switch(self.ctx, record_dict)
        count = self.ctx.session.query(models.PhysicalSwitches).count()
        self.assertEqual(count, 0)

    def _get_physical_port_dict(self):
        uuid = _uuid()
        ps_id = _uuid()
        record_dict = {'uuid': uuid,
                       'name': 'physical_port1',
                       'physical_switch_id': ps_id,
                       'ovsdb_identifier': 'host1'}
        return record_dict

    def _create_physical_port(self,
                              record_dict,
                              name=None,
                              physical_switch_id=None):
        if name and physical_switch_id:
            record_dict['name'] = name
            record_dict['physical_switch_id'] = physical_switch_id
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = models.PhysicalPorts(
                uuid=record_dict['uuid'],
                name=record_dict['name'],
                physical_switch_id=record_dict['physical_switch_id'],
                ovsdb_identifier=record_dict['ovsdb_identifier'])
            self.ctx.session.add(entry)
            return entry

    def test_get_physical_port(self):
        record_dict = self._get_physical_port_dict()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_physical_port(record_dict)
        result = lib.get_physical_port(self.ctx, record_dict)
        self._assert_model_equals(entry, result)

    def test_add_physical_port(self):
        record_dict = self._get_physical_port_dict()
        self._create_physical_port(record_dict)
        count = self.ctx.session.query(models.PhysicalPorts).count()
        self.assertEqual(1, count)

    def test_add_physical_port_raise_on_duplicate_constraint(self):
        record_dict = self._get_physical_port_dict()
        self._create_physical_port(record_dict)
        # Call the method twice to trigger a db duplicate constraint error,
        # this time with a different switch name and physical switch id!
        self.assertRaises(d_exc.DBDuplicateEntry,
                          self._create_physical_port,
                          record_dict, 'physical_port2', _uuid())

    def test_delete_physical_port(self):
        record_dict = self._get_physical_port_dict()
        self._create_physical_port(record_dict)
        lib.delete_physical_port(self.ctx, record_dict)
        count = self.ctx.session.query(models.PhysicalPorts).count()
        self.assertEqual(count, 0)

    def _get_ucast_mac_local_dict(self):
        uuid = _uuid()
        ls_id = _uuid()
        pl_id = _uuid()
        record_dict = {'uuid': uuid,
                       'mac': '12:34:56:78:90:aa:bb',
                       'logical_switch_id': ls_id,
                       'physical_locator_id': pl_id,
                       'ip_address': '10.0.0.1',
                       'ovsdb_identifier': 'host1'}
        return record_dict

    def _create_ucast_mac_local(self, record_dict):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = models.UcastMacsLocals(
                uuid=record_dict['uuid'],
                mac=record_dict['mac'],
                logical_switch_id=record_dict['logical_switch_id'],
                physical_locator_id=record_dict['physical_locator_id'],
                ip_address=record_dict['ip_address'],
                ovsdb_identifier=record_dict['ovsdb_identifier'])
            self.ctx.session.add(entry)
            return entry

    def test_get_ucast_mac_local(self):
        record_dict = self._get_ucast_mac_local_dict()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_ucast_mac_local(record_dict)
        result = lib.get_ucast_mac_local(self.ctx, record_dict)
        self._assert_model_equals(entry, result)

    def test_add_ucast_mac_local(self):
        record_dict = self._get_ucast_mac_local_dict()
        self._create_ucast_mac_local(record_dict)
        count = self.ctx.session.query(models.UcastMacsLocals).count()
        self.assertEqual(1, count)

    def test_add_ucast_mac_local_raise_on_duplicate_constraint(self):
        record_dict = self._get_ucast_mac_local_dict()
        self._create_ucast_mac_local(record_dict)
        # Call the method twice to trigger a db duplicate constraint error,
        # this time with a different mac and logical switch id!
        record_dict['mac'] = '11:22:33:44:55:66:77'
        record_dict['logical_switch_id'] = _uuid()
        self.assertRaises(d_exc.DBDuplicateEntry,
                          self._create_ucast_mac_local,
                          record_dict)

    def test_delete_ucast_mac_local(self):
        record_dict = self._get_ucast_mac_local_dict()
        self._create_ucast_mac_local(record_dict)
        lib.delete_ucast_mac_local(self.ctx, record_dict)
        count = self.ctx.session.query(models.UcastMacsLocals).count()
        self.assertEqual(count, 0)

    def _get_ucast_mac_remote_dict(self):
        uuid = _uuid()
        ls_id = _uuid()
        pl_id = _uuid()
        record_dict = {'uuid': uuid,
                       'mac': '12:34:56:78:90:aa:bb',
                       'logical_switch_id': ls_id,
                       'physical_locator_id': pl_id,
                       'ip_address': '10.0.0.1',
                       'ovsdb_identifier': 'host1'}
        return record_dict

    def _create_ucast_mac_remote(self,
                                 record_dict,
                                 mac=None,
                                 logical_switch_uuid=None):
        if mac and logical_switch_uuid:
            record_dict['mac'] = mac
            record_dict['logical_switch_id'] = logical_switch_uuid
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = models.UcastMacsRemotes(
                uuid=record_dict['uuid'],
                mac=record_dict['mac'],
                logical_switch_id=record_dict['logical_switch_id'],
                physical_locator_id=record_dict['physical_locator_id'],
                ip_address=record_dict['ip_address'],
                ovsdb_identifier=record_dict['ovsdb_identifier'])
            self.ctx.session.add(entry)
            return entry

    def test_get_ucast_mac_remote(self):
        record_dict = self._get_ucast_mac_remote_dict()
        entry = self._create_ucast_mac_remote(record_dict)
        result = lib.get_ucast_mac_remote(self.ctx, record_dict)
        for ent_key, ent_val in entry.items():
            self.assertEqual(ent_val, result[ent_key])

    def test_add_ucast_mac_remote(self):
        record_dict = self._get_ucast_mac_remote_dict()
        self._create_ucast_mac_remote(record_dict)
        count = self.ctx.session.query(models.UcastMacsRemotes).count()
        self.assertEqual(1, count)

    def test_add_ucast_mac_remote_raise_on_duplicate_constraint(self):
        record_dict = self._get_ucast_mac_remote_dict()
        self._create_ucast_mac_remote(record_dict)
        # Call the method twice to trigger a db duplicate constraint error,
        # this time with a different mac and logical switch id!
        self.assertRaises(d_exc.DBDuplicateEntry,
                          self._create_ucast_mac_remote,
                          record_dict, '11:22:33:44:55:66:77', _uuid())

    def test_delete_ucast_mac_remote(self):
        record_dict = self._get_ucast_mac_remote_dict()
        self._create_ucast_mac_remote(record_dict)
        lib.delete_ucast_mac_remote(self.ctx, record_dict)
        count = self.ctx.session.query(models.UcastMacsRemotes).count()
        self.assertEqual(count, 0)

    def _get_vlan_binding_dict(self):
        port_uuid = _uuid()
        ls_uuid = _uuid()
        record_dict = {'port_uuid': port_uuid,
                       'vlan': 200,
                       'logical_switch_uuid': ls_uuid,
                       'ovsdb_identifier': 'host1'}
        return record_dict

    def _get_vlan0_binding_dict(self):
        port_uuid = _uuid()
        ls_uuid = _uuid()
        record_dict = {'port_uuid': port_uuid,
                       'vlan': 0,
                       'logical_switch_uuid': ls_uuid,
                       'ovsdb_identifier': 'host1'}
        return record_dict

    def _create_vlan_binding(self, record_dict, port_uuid=None):
        if port_uuid:
            record_dict['port_uuid'] = port_uuid
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = models.VlanBindings(
                port_uuid=record_dict['port_uuid'],
                vlan=record_dict['vlan'],
                logical_switch_uuid=record_dict['logical_switch_uuid'],
                ovsdb_identifier=record_dict['ovsdb_identifier'])
            self.ctx.session.add(entry)
            return entry

    def test_get_vlan_binding(self):
        record_dict = self._get_vlan_binding_dict()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_vlan_binding(record_dict)
        result = lib.get_vlan_binding(self.ctx, record_dict)
        self._assert_model_equals(entry, result)

    def test_add_vlan_binding(self):
        record_dict = self._get_vlan_binding_dict()
        self._create_vlan_binding(record_dict)
        count = self.ctx.session.query(models.VlanBindings).count()
        self.assertEqual(1, count)

    def test_add_vlan_binding_raise_on_duplicate_constraint(self):
        record_dict = self._get_vlan_binding_dict()
        self._create_vlan_binding(record_dict)
        # Call the method twice to trigger a db duplicate constraint error,
        # this time with a same entries
        self.assertRaises(d_exc.DBDuplicateEntry,
                          self._create_vlan_binding,
                          record_dict)

    def test_delete_vlan_binding(self):
        record_dict = self._get_vlan_binding_dict()
        self._create_vlan_binding(record_dict)
        lib.delete_vlan_binding(self.ctx, record_dict)
        count = self.ctx.session.query(models.VlanBindings).count()
        self.assertEqual(count, 0)

    def test_delete_vlan0_binding(self):
        record_dict = self._get_vlan0_binding_dict()
        self._create_vlan_binding(record_dict)
        lib.delete_vlan_binding(self.ctx, record_dict)
        count = self.ctx.session.query(models.VlanBindings).count()
        self.assertEqual(count, 0)

    def test_get_logical_switch_by_name(self):
        record_dict = self._get_logical_switch_dict()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_logical_switch(record_dict,
                                                'logical_switch2')
        record_dict['logical_switch_name'] = 'logical_switch2'
        result = lib.get_logical_switch_by_name(self.ctx, record_dict)
        self._assert_model_equals(entry, result)

    def test_get_all_logical_switch_by_name(self):
        record_dict1 = self._get_logical_switch_dict()
        self._create_logical_switch(record_dict1, 'logical_switch2')
        record_dict2 = self._get_logical_switch_dict()
        self._create_logical_switch(record_dict2, 'logical_switch2')
        ls_list = lib.get_all_logical_switches_by_name(self.ctx,
                                                       'logical_switch2')
        self.assertEqual(2, len(ls_list))

    def test_get_physical_locator_by_dst_ip(self):
        record_dict = self._get_physical_locator_dict()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_physical_locator(record_dict, '20.0.0.1')
        record_dict['dst_ip'] = '20.0.0.1'
        record_dict.pop('uuid')
        result = lib.get_physical_locator_by_dst_ip(self.ctx, record_dict)
        self._assert_model_equals(entry, result)

    def test_get_physical_switch_by_name(self):
        record_dict = self._get_physical_switch_dict()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_physical_switch(record_dict,
                                                 'physical_switch2')
        result = lib.get_physical_switch_by_name(self.ctx,
                                                 'physical_switch2')
        self._assert_model_equals(entry, result)

    def test_get_all_vlan_bindings_by_physical_port(self):
        record_dict1 = {'port_uuid': 'ps123',
                        'vlan': 200,
                        'logical_switch_uuid': 'ls123',
                        'ovsdb_identifier': 'host1'}
        self._create_vlan_binding(record_dict1)
        record_dict1['vlan'] = 300
        record_dict1['logical_switch_uuid'] = 'ls456'
        self._create_vlan_binding(record_dict1)
        record_dict1['uuid'] = record_dict1.get('port_uuid')
        vlan_list = lib.get_all_vlan_bindings_by_physical_port(self.ctx,
                                                               record_dict1)
        self.assertEqual(2, len(vlan_list))

    def test_get_ucast_mac_remote_by_mac_and_ls(self):
        record_dict = self._get_ucast_mac_remote_dict()
        entry = self._create_ucast_mac_remote(record_dict,
                                              '00:11:22:33:44:55:66',
                                              'ls123')
        record_dict['mac'] = '00:11:22:33:44:55:66'
        record_dict['logical_switch_uuid'] = 'ls123'
        result = lib.get_ucast_mac_remote_by_mac_and_ls(self.ctx,
                                                        record_dict)
        for ent_key, ent_val in entry.items():
            self.assertEqual(ent_val, result[ent_key])

    def test_get_ucast_mac_remote_by_mac_and_ls_when_not_found(self):
        record_dict = self._get_ucast_mac_remote_dict()
        record_dict['mac'] = '00:11:22:33:44:55:66'
        record_dict['logical_switch_uuid'] = 'ls123'
        result = lib.get_ucast_mac_remote_by_mac_and_ls(self.ctx,
                                                        record_dict)
        self.assertIsNone(result)

    def test_get_physical_port_by_name_and_ps(self):
        record_dict = self._get_physical_port_dict()
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_physical_port(record_dict,
                                               'port1',
                                               'ps123')
        record_dict['interface_name'] = 'port1'
        record_dict['physical_switch_id'] = 'ps123'
        result = lib.get_physical_port_by_name_and_ps(self.ctx,
                                                      record_dict)
        self._assert_model_equals(entry, result)

    def _get_pending_mac_dict(self, timestamp):
        record_dict = {'uuid': _uuid(),
                       'mac': 'aa:aa:aa:aa:aa:aa',
                       'logical_switch_uuid': 'fake_ls_id',
                       'locator_uuid': _uuid(),
                       'dst_ip': '1.1.1.1',
                       'vm_ip': '2.2.2.2',
                       'ovsdb_identifier': 'ovsdb1',
                       'operation': 'insert',
                       'timestamp': timestamp}
        return record_dict

    def _create_pending_mac(self, record_dict):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = models.PendingUcastMacsRemote(
                uuid=record_dict['uuid'],
                mac=record_dict['mac'],
                logical_switch_uuid=record_dict['logical_switch_uuid'],
                locator_uuid=record_dict['locator_uuid'],
                dst_ip=record_dict['dst_ip'],
                vm_ip=record_dict['vm_ip'],
                ovsdb_identifier=record_dict['ovsdb_identifier'],
                operation=record_dict['operation'],
                timestamp=record_dict['timestamp'])
            self.ctx.session.add(entry)
            return entry

    def test_add_pending_ucast_mac_remote(self):
        timestamp = timeutils.utcnow()
        record_dict = self._get_pending_mac_dict(timestamp)
        self._create_pending_mac(record_dict)
        count = self.ctx.session.query(models.PendingUcastMacsRemote).count()
        self.assertEqual(1, count)

    def test_get_pending_ucast_mac_remote(self):
        timestamp = timeutils.utcnow()
        record_dict = self._get_pending_mac_dict(timestamp)
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry = self._create_pending_mac(record_dict)
        result = lib.get_pending_ucast_mac_remote(
            self.ctx, record_dict['ovsdb_identifier'],
            record_dict['mac'], record_dict['logical_switch_uuid'])
        self._assert_model_equals(entry, result)

    def test_get_all_pending_remote_macs_in_asc_order(self):
        timestamp1 = timeutils.utcnow()
        record_dict1 = self._get_pending_mac_dict(timestamp1)
        timestamp2 = timeutils.utcnow()
        record_dict2 = self._get_pending_mac_dict(timestamp2)
        entry = []
        with db_api.CONTEXT_WRITER.using(self.ctx):
            entry.append(self._create_pending_mac(record_dict1))
            entry.append(self._create_pending_mac(record_dict2))
        result = lib.get_all_pending_remote_macs_in_asc_order(
            self.ctx, record_dict1['ovsdb_identifier'])
        for index, res in enumerate(result):
            for k, v in res.items():
                self.assertEqual(res[k], entry[index][k])

    def test_delete_pending_ucast_mac_remote(self):
        timestamp = timeutils.utcnow()
        record_dict = self._get_pending_mac_dict(timestamp)
        self._create_pending_mac(record_dict)
        lib.delete_pending_ucast_mac_remote(self.ctx,
                                            record_dict['operation'],
                                            record_dict['ovsdb_identifier'],
                                            record_dict['logical_switch_uuid'],
                                            record_dict['mac'])
        count = self.ctx.session.query(models.UcastMacsRemotes).count()
        self.assertEqual(count, 0)

    def test_get_all_ucast_mac_remote_by_ls(self):
        record_dict = self._get_ucast_mac_remote_dict()
        record_dict1 = self._create_ucast_mac_remote(record_dict)
        record_dict = self._get_ucast_mac_remote_dict()
        record_dict['mac'] = '00:11:22:33:44:55:66'
        record_dict['logical_switch_id'] = record_dict1.get(
            'logical_switch_id')
        self._create_ucast_mac_remote(record_dict)
        mac_list = lib.get_all_ucast_mac_remote_by_ls(self.ctx, record_dict)
        self.assertEqual(2, len(mac_list))



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.common import config as common_config
from neutron.tests import base

from oslo_config import cfg

from networking_l2gw.services.l2gateway.common import config
from networking_l2gw.services.l2gateway import l2gw_agent as agent


class TestL2gwAgent(base.BaseTestCase):
    def setUp(self):
        super(TestL2gwAgent, self).setUp()
        config.register_ovsdb_opts_helper(cfg.CONF)

    def test_start(self):
        with mock.patch.object(
            agent.n_rpc.Service, 'start'
        ) as mock_start:
            mgr = mock.Mock()
            cfg.CONF.periodic_interval = mock.Mock(return_value=10)
            agent_service = agent.L2gatewayAgentService('host',
                                                        'topic', mgr)
            agent_service.start()
            self.assertTrue(mock_start.called)

    def test_main_l2gw_agent(self):
        logging_str = 'neutron.conf.agent.common.setup_logging'
        common_config_str = mock.patch.object(common_config, 'init').start()
        with mock.patch.object(common_config_str, 'init'), \
                mock.patch(logging_str), \
                mock.patch.object(agent.service, 'launch') as mock_launch, \
                mock.patch('sys.argv'), \
                mock.patch.object(agent.manager, 'OVSDBManager'), \
                mock.patch.object(cfg.CONF, 'register_opts'):
            agent.main()
            mock_launch.assert_called_once_with(cfg.CONF, mock.ANY)



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os.path
import socket
import ssl
import time

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils

from networking_l2gw.services.l2gateway.common import constants as n_const

LOG = logging.getLogger(__name__)
OVSDB_UNREACHABLE_MSG = 'Unable to reach OVSDB server %s'
OVSDB_CONNECTED_MSG = 'Connected to OVSDB server %s'


class BaseConnection(object):
    """Connects to OVSDB server.

       Connects to an ovsdb server with/without SSL
       on a given host and TCP port.
    """
    def __init__(self, conf, gw_config, mgr=None):
        self.responses = []
        self.connected = False
        self.mgr = mgr
        self.enable_manager = cfg.CONF.ovsdb.enable_manager
        if self.enable_manager:
            self.manager_table_listening_port = (
                cfg.CONF.ovsdb.manager_table_listening_port)
            self.ip_ovsdb_mapping = self._get_ovsdb_ip_mapping()
            self.s = None
            self.check_c_sock = None
            self.check_sock_rcv = False
            self.ovsdb_dicts = {}
            self.ovsdb_fd_states = {}
            self.ovsdb_conn_list = []
            eventlet.greenthread.spawn(self._rcv_socket)
        else:
            self.gw_config = gw_config
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if gw_config.use_ssl:
                ssl_sock = ssl.wrap_socket(
                    self.socket,
                    server_side=False,
                    keyfile=gw_config.private_key,
                    certfile=gw_config.certificate,
                    cert_reqs=ssl.CERT_REQUIRED,
                    ssl_version=ssl.PROTOCOL_TLSv1,
                    ca_certs=gw_config.ca_cert)
                self.socket = ssl_sock
            retryCount = 0
            while True:
                try:
                    self.socket.connect((str(gw_config.ovsdb_ip),
                                         int(gw_config.ovsdb_port)))
                    break
                except (socket.error, socket.timeout):
                    LOG.warning(OVSDB_UNREACHABLE_MSG, gw_config.ovsdb_ip)
                    if retryCount == conf.max_connection_retries:
                        # Retried for max_connection_retries times.
                        # Give up and return so that it can be tried in
                        # the next periodic interval.
                        with excutils.save_and_reraise_exception(reraise=True):
                            LOG.exception("Socket error in connecting to "
                                          "the OVSDB server")
                    else:
                        time.sleep(1)
                        retryCount += 1

            # Successfully connected to the socket
            LOG.debug(OVSDB_CONNECTED_MSG, gw_config.ovsdb_ip)
            self.connected = True

    def _get_ovsdb_ip_mapping(self):
        ovsdb_ip_mapping = {}
        ovsdb_hosts = cfg.CONF.ovsdb.ovsdb_hosts
        if ovsdb_hosts != '':
            ovsdb_hosts = ovsdb_hosts.split(',')
            for host in ovsdb_hosts:
                host_splits = str(host).split(':')
                ovsdb_identifier = str(host_splits[0]).strip()
                ovsdb_ip = str(host_splits[1]).strip()
                ovsdb_ip_mapping[ovsdb_ip] = ovsdb_identifier
            return ovsdb_ip_mapping

    def _is_ssl_configured(self, addr, client_sock):
        priv_key_path = cfg.CONF.ovsdb.l2_gw_agent_priv_key_base_path
        cert_path = cfg.CONF.ovsdb.l2_gw_agent_cert_base_path
        ca_cert_path = cfg.CONF.ovsdb.l2_gw_agent_ca_cert_base_path
        use_ssl = priv_key_path and cert_path and ca_cert_path
        if use_ssl:
            LOG.debug("ssl is enabled with priv_key_path %s, cert_path %s, "
                      "ca_cert_path %s", priv_key_path,
                      cert_path, ca_cert_path)
            if addr in self.ip_ovsdb_mapping.keys():
                ovsdb_id = self.ip_ovsdb_mapping.get(addr)
                priv_key_file = priv_key_path + "/" + ovsdb_id + ".key"
                cert_file = cert_path + "/" + ovsdb_id + ".cert"
                ca_cert_file = ca_cert_path + "/" + ovsdb_id + ".ca_cert"
                is_priv_key = os.path.isfile(priv_key_file)
                is_cert_file = os.path.isfile(cert_file)
                is_ca_cert_file = os.path.isfile(ca_cert_file)
                if is_priv_key and is_cert_file and is_ca_cert_file:
                    ssl_conn_stream = ssl.wrap_socket(
                        client_sock,
                        server_side=True,
                        keyfile=priv_key_file,
                        certfile=cert_file,
                        ssl_version=ssl.PROTOCOL_SSLv23,
                        ca_certs=ca_cert_file)
                    client_sock = ssl_conn_stream
                else:
                    if not is_priv_key:
                        LOG.error("Could not find private key in"
                                  " %(path)s dir, expecting in the "
                                  "file name %(file)s ",
                                  {'path': priv_key_path,
                                   'file': ovsdb_id + ".key"})
                    if not is_cert_file:
                        LOG.error("Could not find cert in %(path)s dir, "
                                  "expecting in the file name %(file)s",
                                  {'path': cert_path,
                                   'file': ovsdb_id + ".cert"})
                    if not is_ca_cert_file:
                        LOG.error("Could not find cacert in %(path)s "
                                  "dir, expecting in the file name "
                                  "%(file)s",
                                  {'path': ca_cert_path,
                                   'file': ovsdb_id + ".ca_cert"})
            else:
                LOG.error("you have enabled SSL for ovsdb %s, "
                          "expecting the ovsdb identifier and ovsdb IP "
                          "entry in ovsdb_hosts in l2gateway_agent.ini",
                          addr)
        return client_sock

    def _rcv_socket(self):
        # Create a socket object.
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        host = ''                        # Get local machine name
        port = self.manager_table_listening_port
        # configured port for your service.
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.s.bind((host, port))        # Bind to the port
        self.s.listen(5)                 # Now wait for client connection.
        while True:
            # Establish connection with client.
            c_sock, ip_addr = self.s.accept()
            addr = ip_addr[0]
            c_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            c_sock = self._is_ssl_configured(addr, c_sock)
            LOG.debug("Got connection from %s ", addr)
            self.connected = True
            if addr in self.ovsdb_fd_states.keys():
                del self.ovsdb_fd_states[addr]
            if addr in self.ovsdb_conn_list:
                self.ovsdb_conn_list.remove(addr)
            if addr in self.ovsdb_dicts.keys():
                if self.ovsdb_dicts.get(addr):
                    self.ovsdb_dicts.get(addr).close()
                del self.ovsdb_dicts[addr]
            self.ovsdb_dicts[addr] = c_sock
            eventlet.greenthread.spawn(self._common_sock_rcv_thread, addr)
            # Now that OVSDB server has sent a socket open request, let us wait
            # for echo request. After the first echo request, we will send the
            # "monitor" request to the OVSDB server.

    def _send_monitor_msg_to_ovsdb_connection(self, addr):
        if self.mgr.l2gw_agent_type == n_const.MONITOR:
            try:
                if (self.mgr.ovsdb_fd) and (addr in self.ovsdb_conn_list):
                    eventlet.greenthread.spawn_n(
                        self.mgr.ovsdb_fd._spawn_monitor_table_thread,
                        addr)
            except Exception:
                LOG.warning("Could not send monitor message to the "
                            "OVSDB server.")
                self.disconnect(addr)

    def _common_sock_rcv_thread(self, addr):
        chunks = []
        lc = rc = 0
        prev_char = None
        self.read_on = True
        check_monitor_msg = True
        self._echo_response(addr)
        if self.enable_manager and (addr in self.ovsdb_conn_list):
            while self.read_on:
                response = self.ovsdb_dicts.get(addr).recv(n_const.BUFFER_SIZE)
                self.ovsdb_fd_states[addr] = 'connected'
                self.check_sock_rcv = True
                eventlet.greenthread.sleep(0)
                if check_monitor_msg:
                    self._send_monitor_msg_to_ovsdb_connection(addr)
                    check_monitor_msg = False
                if response:
                    response = response.decode('utf8')
                    message_mark = 0
                    for i, c in enumerate(response):
                        if c == '{' and not (prev_char and
                                             prev_char == '\\'):
                            lc += 1
                        elif c == '}' and not (prev_char and
                                               prev_char == '\\'):
                            rc += 1
                        if rc > lc:
                            raise Exception("json string not valid")
                        elif lc == rc and lc != 0:
                            chunks.append(response[message_mark:i + 1])
                            message = "".join(chunks)
                            eventlet.greenthread.spawn_n(
                                self._on_remote_message, message, addr)
                            eventlet.greenthread.sleep(0)
                            lc = rc = 0
                            message_mark = i + 1
                            chunks = []
                        prev_char = c
                    chunks.append(response[message_mark:])
                else:
                    self.read_on = False
                    self.disconnect(addr)

    def _echo_response(self, addr):
        while True:
            try:
                if self.enable_manager:
                    eventlet.greenthread.sleep(0)
                    response = self.ovsdb_dicts.get(addr).recv(
                        n_const.BUFFER_SIZE)
                    sock_json_m = jsonutils.loads(response)
                    sock_handler_method = sock_json_m.get('method', None)
                    if sock_handler_method == 'echo':
                        self.check_c_sock = True
                        self.ovsdb_dicts.get(addr).send(jsonutils.dumps(
                            {"result": sock_json_m.get("params", None),
                             "error": None, "id": sock_json_m['id']}))
                        if (addr not in self.ovsdb_conn_list):
                            self.ovsdb_conn_list.append(addr)
                        break
            except Exception:
                continue

    def send(self, message, callback=None, addr=None):
        """Sends a message to the OVSDB server."""
        if callback:
            self.callbacks[message['id']] = callback
        retry_count = 0
        bytes_sent = 0
        while retry_count <= n_const.MAX_RETRIES:
            try:
                data = jsonutils.dumps(message)
                if not isinstance(data, bytes):
                    data = data.encode()
                if self.enable_manager:
                    bytes_sent = self.ovsdb_dicts.get(addr).send(data)
                else:
                    bytes_sent = self.socket.send(data)
                if bytes_sent:
                    return True
            except Exception as ex:
                LOG.exception("Exception [%s] occurred while sending "
                              "message to the OVSDB server", ex)
            retry_count += 1

        LOG.warning("Could not send message to the "
                    "OVSDB server.")
        self.disconnect(addr)
        return False

    def disconnect(self, addr=None):
        """disconnects the connection from the OVSDB server."""
        if self.enable_manager:
            self.ovsdb_dicts.get(addr).close()
            del self.ovsdb_dicts[addr]
            if addr in self.ovsdb_fd_states.keys():
                del self.ovsdb_fd_states[addr]
            self.ovsdb_conn_list.remove(addr)
        else:
            self.socket.close()
        self.connected = False

    def _response(self, operation_id):
        x_copy = None
        to_delete = None
        for x in self.responses:
            if x['id'] == operation_id:
                x_copy = copy.deepcopy(x)
                to_delete = x
                break
        if to_delete:
            self.responses.remove(to_delete)
        return x_copy



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""l2gateway_models

Revision ID: 42438454c556
Revises: 54c9c8fe22bf
Create Date: 2014-11-27 01:57:56.997665

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '42438454c556'
down_revision = '54c9c8fe22bf'


def upgrade():
    op.create_table('l2gateways',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('name', sa.String(length=255), nullable=True),
                    sa.Column('tenant_id', sa.String(length=255),
                              nullable=True),
                    sa.PrimaryKeyConstraint('id'))

    op.create_table('l2gatewaydevices',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('device_name', sa.String(length=255),
                              nullable=False),
                    sa.Column('l2_gateway_id', sa.String(length=36),
                              nullable=False),
                    sa.ForeignKeyConstraint(['l2_gateway_id'],
                                            ['l2gateways.id'],
                                            ondelete='CASCADE'),
                    sa.PrimaryKeyConstraint('id'))

    op.create_table('l2gatewayinterfaces',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('interface_name', sa.String(length=255),
                              nullable=True),
                    sa.Column('segmentation_id', sa.Integer(),
                              nullable=True),
                    sa.Column('device_id', sa.String(length=36),
                              nullable=False),
                    sa.ForeignKeyConstraint(['device_id'],
                                            ['l2gatewaydevices.id'],
                                            ondelete='CASCADE'),
                    sa.PrimaryKeyConstraint('id'))

    op.create_table('l2gatewayconnections',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('tenant_id', sa.String(length=255),
                              nullable=True),
                    sa.Column('l2_gateway_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('network_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('segmentation_id', sa.Integer(),
                              nullable=True),
                    sa.ForeignKeyConstraint(['l2_gateway_id'],
                                            ['l2gateways.id'],
                                            ondelete='CASCADE'),
                    sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                                            ondelete='CASCADE'),
                    sa.UniqueConstraint('l2_gateway_id',
                                        'network_id'),
                    sa.PrimaryKeyConstraint('id'))

    op.create_table('pending_ucast_macs_remotes',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('uuid', sa.String(length=36), nullable=True),
                    sa.Column('mac', sa.String(32), nullable=False),
                    sa.Column('logical_switch_uuid', sa.String(36),
                              nullable=False),
                    sa.Column('locator_uuid', sa.String(36),
                              nullable=True),
                    sa.Column('dst_ip', sa.String(64)),
                    sa.Column('vm_ip', sa.String(64)),
                    sa.Column('ovsdb_identifier', sa.String(64),
                              nullable=False),
                    sa.Column('operation', sa.String(8), nullable=False),
                    sa.Column('timestamp', sa.DateTime, nullable=False))



# Copyright 2015 OpenStack Foundation
# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.api import faults
from neutron_lib import exceptions
from webob import exc as web_exc

from networking_l2gw._i18n import _


class L2GatewayInUse(exceptions.InUse):
    message = _("L2 Gateway '%(gateway_id)s' still has active mappings "
                "with one or more neutron networks.")


class L2GatewayNotFound(exceptions.NotFound):
    message = _("L2 Gateway %(gateway_id)s could not be found")


class L2GatewayDeviceInUse(exceptions.InUse):
    message = _("L2 Gateway Device '%(device_id)s' is still used by "
                "one or more network gateways.")


class L2AgentNotFoundByHost(exceptions.NotFound):
    message = _("L2 Agent for host '%(host)s' could not be found.")


class L2GatewayDeviceNotFound(exceptions.NotFound):
    message = _("L2 Gateway Device %(device_id)s could not be found.")


class L2GatewayDeviceNameNotFound(exceptions.NotFound):
    message = _("L2 Gateway Device %(device_name)s could not be found.")


class L2GatewayPortInUse(exceptions.InUse):
    message = _("Port '%(port_id)s' is owned by '%(device_owner)s' and "
                "therefore cannot be deleted directly via the port API.")


class L2GatewayConnectionExists(exceptions.InUse):
    message = _("The specified mapping '%(mapping)s' exists on "
                "network gateway '%(gateway_id)s'.")


class L2MultipleGatewayConnections(exceptions.NeutronException):
    message = _("Multiple network connections found on '%(gateway_id)s' "
                "with provided criteria.")


class L2GatewayInterfaceNotFound(exceptions.NeutronException):
    message = _("L2 Gateway interface '%(interface_id)s' not found.")


class L2GatewayPhysicalPortNotFound(exceptions.NeutronException):
    message = _("Physical Port '%(int_name)s' in Physical switch "
                "'%(device_name)s' could not be found")


class L2GatewayPhysicalPortFaultStatus(exceptions.NeutronException):
    message = _("Physical Port '%(int_name)s' in Physical switch "
                "'%(device_name)s' port_fault_status is '%(fault_status)s'")


class L2GatewayPhysicalSwitchFaultStatus(exceptions.NeutronException):
    message = _("Physical Switch '%(device_name)s' switch_fault_status is "
                "'%(fault_status)s'")


class L2GatewayConnectionNotFound(exceptions.NotFound):
    message = _("The connection %(id)s not found on the l2 gateway")


class L2gatewaySegmentationIDNotFound(exceptions.NotFound):
    message = _("The segmentation id not found on gateway "
                "'%(gateway_id)s'")


class MultipleVxlanSegmentsFound(exceptions.NeutronException):
    message = _("Multiple Vxlan segments found for the network "
                "'%(network_id)s'")


class VxlanSegmentationIDNotFound(exceptions.NotFound):
    message = _("vxlan segmentation id not found for the "
                "network '%(network_id)s'")


class L2GatewayInterfaceRequired(exceptions.NeutronException):
    message = _("L2 Gateway Interface required")


class L2GatewaySegmentationIDExists(exceptions.NeutronException):
    message = _("The segmentation id is already specified in "
                "existing gateway interfaces.")


class L2GatewaySegmentationIDNotExists(exceptions.NeutronException):
    message = _("The segmentation id is not specified in "
                "existing gateway interfaces.")


class L2GatewaySegmentationRequired(exceptions.NeutronException):
    message = _("L2 gateway segmentation id must be consistent for all "
                "the interfaces")


class L2GatewayDuplicateSegmentationID(exceptions.Conflict):
    message = _("%(message)s")


class OVSDBError(exceptions.NeutronException):
    message = _("%(message)s")


class L2GatewayServiceDriverError(exceptions.NeutronException):
    """Service driver call failed."""
    message = _("%(method)s failed.")


class InvalidMethod(exceptions.NeutronException):
    message = _("invalid method '%(op_method)s'")


faults.FAULT_MAP.update({L2GatewayInUse: web_exc.HTTPConflict,
                         L2GatewayPortInUse: web_exc.HTTPConflict,
                         L2GatewayConnectionExists: web_exc.HTTPConflict,
                         L2GatewayConnectionNotFound: web_exc.HTTPNotFound,
                         MultipleVxlanSegmentsFound: web_exc.HTTPConflict,
                         VxlanSegmentationIDNotFound: web_exc.HTTPNotFound,
                         L2GatewaySegmentationRequired: web_exc.HTTPConflict,
                         L2MultipleGatewayConnections: web_exc.HTTPConflict,
                         L2GatewayDuplicateSegmentationID:
                             web_exc.HTTPConflict,
                         L2AgentNotFoundByHost: web_exc.HTTPNotFound,
                         OVSDBError: web_exc.HTTPConflict})


class L3DvrAgentNotFound(exceptions.NotFound):
    message = _("L3 agent could not be found")


class DvrAgentHostnameNotFound(exceptions.NeutronException):
    message = _("Hostname '%(host)' has 127.0.0.1 address")



# All Rights Reserved 2018
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import operator
from unittest import mock

from osc_lib import utils as osc_utils
from osc_lib.utils import columns as column_util

from neutronclient.tests.unit.osc.v2 import fakes as test_fakes

from networking_l2gw.l2gatewayclient.osc import l2gw_connection as \
    osc_l2gw_conn
from networking_l2gw.tests.unit.l2gatewayclient.osc import fakes

columns_long = tuple(col for col, _, listing_mode in osc_l2gw_conn._attr_map
                     if listing_mode in (column_util.LIST_BOTH,
                                         column_util.LIST_LONG_ONLY))
headers_long = tuple(head for _, head, listing_mode in osc_l2gw_conn._attr_map
                     if listing_mode in (column_util.LIST_BOTH,
                                         column_util.LIST_LONG_ONLY))
sorted_attr_map = sorted(osc_l2gw_conn._attr_map, key=operator.itemgetter(1))
sorted_columns = tuple(col for col, _, _ in sorted_attr_map)
sorted_headers = tuple(head for _, head, _ in sorted_attr_map)


def _get_data(attrs, columns=sorted_columns):
    return osc_utils.get_dict_properties(attrs, columns)


class TestCreateL2gwConnection(test_fakes.TestNeutronClientOSCV2):
    columns = (
        'ID',
        'L2 GateWay ID',
        'Network ID',
        'Segmentation ID',
        'Tenant'
    )

    def setUp(self):
        super(TestCreateL2gwConnection, self).setUp()
        self.cmd = osc_l2gw_conn.CreateL2gwConnection(
            self.app, self.namespace)
        self.neutronclient.find_resource = mock.Mock(
            side_effect=lambda _, name_or_id, *x: {'id': name_or_id})

    def test_create_l2gateway_connection(self):
        """Test Create l2gateway-connection."""

        fake_connection = fakes.FakeL2GWConnection.create_l2gw_connection()
        self.neutronclient.post = mock.Mock(
            return_value={
                osc_l2gw_conn.L2_GATEWAY_CONNECTION: fake_connection
            })

        arg_list = [
            fake_connection['l2_gateway_id'],
            fake_connection['network_id'],
            '--default-segmentation-id', fake_connection['segmentation_id']
        ]
        verify_list = [
            ('gateway_name', fake_connection['l2_gateway_id']),
            ('network', fake_connection['network_id']),
            ('seg_id', fake_connection['segmentation_id'])
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.neutronclient.post.assert_called_once_with(
            osc_l2gw_conn.object_path,
            {osc_l2gw_conn.L2_GATEWAY_CONNECTION:
                {'segmentation_id': fake_connection['segmentation_id'],
                 'network_id': fake_connection['network_id'],
                 'l2_gateway_id': fake_connection['l2_gateway_id']
                 }
             }
        )
        self.assertEqual(self.columns, columns)
        self.assertItemEqual(_get_data(fake_connection), data)


class TestListL2gwConnection(test_fakes.TestNeutronClientOSCV2):
    def setUp(self):
        super(TestListL2gwConnection, self).setUp()
        self.cmd = osc_l2gw_conn.ListL2gwConnection(self.app, self.namespace)

    def test_list_l2gateway_connection(self):
        """Test List l2gateway-connections."""

        fake_connections = fakes.FakeL2GWConnection.create_l2gw_connections(
            count=3)
        self.neutronclient.list = mock.Mock(return_value=fake_connections)
        arg_list = []
        verify_list = []

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)

        headers, data = self.cmd.take_action(parsed_args)

        self.neutronclient.list.assert_called_once()
        self.assertEqual(headers, list(headers_long))
        self.assertListItemEqual(
            list(data),
            [_get_data(fake_connection, columns_long) for fake_connection
             in fake_connections[osc_l2gw_conn.L2_GATEWAY_CONNECTIONS]]
        )


class TestShowL2gwConnection(test_fakes.TestNeutronClientOSCV2):
    def setUp(self):
        super(TestShowL2gwConnection, self).setUp()
        self.cmd = osc_l2gw_conn.ShowL2gwConnection(self.app, self.namespace)
        self.neutronclient.find_resource = mock.Mock(
            side_effect=lambda _, name_or_id, *x: {'id': name_or_id})

    def test_show_l2gateway_connection(self):
        """Test Show l2gateway-connection."""

        fake_connection = fakes.FakeL2GWConnection.create_l2gw_connection()
        self.neutronclient.get = mock.Mock(
            return_value={
                osc_l2gw_conn.L2_GATEWAY_CONNECTION: fake_connection
            })
        arg_list = [fake_connection['id']]
        verify_list = [
            (osc_l2gw_conn.L2_GATEWAY_CONNECTION, fake_connection['id'])
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)

        headers, data = self.cmd.take_action(parsed_args)

        self.neutronclient.get.assert_called_once_with(
            osc_l2gw_conn.resource_path % fake_connection['id'])
        self.assertEqual(sorted_headers, headers)
        self.assertItemEqual(_get_data(fake_connection), data)


class TestDeleteL2gwConnection(test_fakes.TestNeutronClientOSCV2):
    def setUp(self):
        super(TestDeleteL2gwConnection, self).setUp()
        self.neutronclient.find_resource = mock.Mock(
            side_effect=lambda _, name_or_id: {'id': name_or_id})
        self.cmd = osc_l2gw_conn.DeleteL2gwConnection(self.app, self.namespace)

    def test_delete_l2gateway_connection(self):
        """Test Delete l2gateway-connection."""

        fake_connection = fakes.FakeL2GWConnection.create_l2gw_connection()
        self.neutronclient.delete = mock.Mock(return_value=fake_connection)
        arg_list = [fake_connection['id']]
        verify_list = [
            (osc_l2gw_conn.L2_GATEWAY_CONNECTIONS, [fake_connection['id']])
        ]

        parsed_args = self.check_parser(self.cmd, arg_list, verify_list)

        result = self.cmd.take_action(parsed_args)

        self.neutronclient.delete.assert_called_once_with(
            osc_l2gw_conn.resource_path % fake_connection['id'])
        self.assertIsNone(result)



# All Rights Reserved 2018
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as osc_utils
from osc_lib.utils import columns as column_util

from neutronclient._i18n import _
from neutronclient.neutron import v2_0 as n_v20
from neutronclient.osc import utils as nc_osc_utils

LOG = logging.getLogger(__name__)

L2_GATEWAY_CONNECTION = 'l2_gateway_connection'
L2_GATEWAY_CONNECTIONS = '%ss' % L2_GATEWAY_CONNECTION

path = 'l2-gateway-connections'
object_path = '/%s' % path
resource_path = '/%s/%%s' % path

_attr_map = (
    ('id', 'ID', column_util.LIST_BOTH),
    ('tenant_id', 'Tenant', column_util.LIST_LONG_ONLY),
    ('l2_gateway_id', 'L2 GateWay ID', column_util.LIST_BOTH),
    ('network_id', 'Network ID', column_util.LIST_BOTH),
    ('segmentation_id', 'Segmentation ID', column_util.LIST_BOTH),
)


class CreateL2gwConnection(command.ShowOne):
    _description = _("Create l2gateway-connection")

    def retrieve_ids(self, client, args):
        gateway_id = n_v20.find_resourceid_by_name_or_id(
            client, 'l2_gateway', args.gateway_name)
        network_id = n_v20.find_resourceid_by_name_or_id(
            client, 'network', args.network)
        return gateway_id, network_id

    def get_parser(self, prog_name):
        parser = super(CreateL2gwConnection, self).get_parser(prog_name)
        parser.add_argument('gateway_name', metavar='<GATEWAY-NAME/UUID>',
                            help=_('Descriptive name for logical gateway.'))
        parser.add_argument(
            'network', metavar='<NETWORK-NAME/UUID>',
            help=_('Network name or uuid.'))
        parser.add_argument(
            '--default-segmentation-id',
            dest='seg_id',
            help=_('default segmentation-id that will '
                   'be applied to the interfaces for which '
                   'segmentation id was not specified '
                   'in l2-gateway-create command.'))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.neutronclient
        (gateway_id, network_id) = self.retrieve_ids(client, parsed_args)
        body = {
            L2_GATEWAY_CONNECTION: {
                'l2_gateway_id': gateway_id, 'network_id': network_id
            }
        }
        if parsed_args.seg_id:
            body[L2_GATEWAY_CONNECTION]['segmentation_id'] = \
                parsed_args.seg_id
        obj = client.post(object_path, body)[L2_GATEWAY_CONNECTION]
        columns, display_columns = column_util.get_columns(obj, _attr_map)
        data = osc_utils.get_dict_properties(obj, columns)
        return display_columns, data


class ListL2gwConnection(command.Lister):
    _description = _("List l2gateway-connections")

    def get_parser(self, prog_name):
        parser = super(ListL2gwConnection, self).get_parser(prog_name)
        nc_osc_utils.add_project_owner_option_to_parser(parser)

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.neutronclient
        params = {}
        if parsed_args.project is not None:
            project_id = nc_osc_utils.find_project(
                self.app.client_manager.identity,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
            params['tenant_id'] = project_id
        objs = client.list(
            L2_GATEWAY_CONNECTIONS, object_path,
            retrieve_all=True, params=params)[L2_GATEWAY_CONNECTIONS]
        headers, columns = column_util.get_column_definitions(
            _attr_map, long_listing=True)
        return (headers, (osc_utils.get_dict_properties(
            s, columns) for s in objs))


class ShowL2gwConnection(command.ShowOne):
    _description = _("Show information of a given l2gateway-connection")

    def get_parser(self, prog_name):
        parser = super(ShowL2gwConnection, self).get_parser(prog_name)
        parser.add_argument(
            L2_GATEWAY_CONNECTION,
            metavar="<L2_GATEWAY_CONNECTION>",
            help=_("ID of l2_gateway_connection to look up."),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.neutronclient
        id = client.find_resource(L2_GATEWAY_CONNECTION,
                                  parsed_args.l2_gateway_connection)['id']
        obj = client.get(resource_path % id)[L2_GATEWAY_CONNECTION]
        columns, display_columns = column_util.get_columns(obj, _attr_map)
        data = osc_utils.get_dict_properties(obj, columns)
        return display_columns, data


class DeleteL2gwConnection(command.Command):
    _description = _("Delete a given l2gateway-connection")

    def get_parser(self, prog_name):
        parser = super(DeleteL2gwConnection, self).get_parser(prog_name)
        parser.add_argument(
            L2_GATEWAY_CONNECTIONS,
            metavar="<L2_GATEWAY_CONNECTIONS>",
            nargs="+",
            help=_("ID(s) of l2_gateway_connections(s) to delete."),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.neutronclient
        fails = 0
        for id_or_name in parsed_args.l2_gateway_connections:
            try:
                id = client.find_resource(
                    L2_GATEWAY_CONNECTION, id_or_name)['id']
                client.delete(resource_path % id)
                LOG.warning("L2 Gateaway Connection %(id)s deleted",
                            {'id': id})
            except Exception as e:
                fails += 1
                LOG.error("Failed to delete L2 Gateway Connection with name "
                          "or ID '%(id_or_name)s': %(e)s",
                          {'id_or_name': id_or_name, 'e': e})
        if fails > 0:
            msg = (_("Failed to delete %(fails)s of %(total)s L2 Gateway "
                     "Connection.") %
                   {'fails': fails, 'total': len(
                       parsed_args.l2_gateway_connections)})
            raise exceptions.CommandError(msg)



# Copyright 2017 <PUT YOUR NAME/COMPANY HERE>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add standard_attribute_id for l2gw

Revision ID: 8d7d772eafcf
Revises: 49ce408ac349
Create Date: 2017-12-19 07:49:23.362289

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '8d7d772eafcf'
down_revision = '49ce408ac349'

TABLES = ("l2gateways", "l2gatewayconnections", "l2gatewaydevices")


def upgrade():
    for table in TABLES:
        op.add_column(table, sa.Column('standard_attr_id', sa.BigInteger(),
                                       nullable=True))



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import rpc as n_rpc

import oslo_messaging as messaging


class L2GatewayAgentApi(object):
    """Agent side of the Agent to Plugin RPC API."""

    API_VERSION = '1.0'

    def __init__(self, topic, host):
        self.host = host
        target = messaging.Target(topic=topic, version=self.API_VERSION)
        self.client = n_rpc.get_client(target)

    def update_ovsdb_changes(self, context, activity, ovsdb_data):
        cctxt = self.client.prepare()
        return cctxt.cast(context,
                          'update_ovsdb_changes',
                          activity=activity,
                          ovsdb_data=ovsdb_data)

    def notify_ovsdb_states(self, context, ovsdb_states):
        cctxt = self.client.prepare()
        return cctxt.cast(context,
                          'notify_ovsdb_states',
                          ovsdb_states=ovsdb_states)
